from ..datatypes import ImpulseResponseFile
from ..datatypes import SourceFile
from ..datatypes import SpectralMask
from ..datatypes import TruthParameter


def raw_load_config(name: str) -> dict:
    """Load YAML config file

    :param name: File name
    :return: Dictionary of config data
    """
    import yaml

    with open(file=name) as f:
        config = yaml.safe_load(f)

    return config


def get_default_config() -> dict:
    """Load default SonusAI config

    :return: Dictionary of default config data
    """
    from .constants import DEFAULT_CONFIG

    try:
        return raw_load_config(DEFAULT_CONFIG)
    except Exception as e:
        raise OSError(f"Error loading default config: {e}") from e


def load_config(name: str) -> dict:
    """Load SonusAI default config and update with given location (performing SonusAI variable substitution)

    :param name: Directory containing mixture database
    :return: Dictionary of config data
    """
    from os.path import join

    return update_config_from_file(filename=join(name, "config.yml"), given_config=get_default_config())


def update_config_from_file(filename: str, given_config: dict) -> dict:
    """Update the given config with the config in the specified YAML file

    :param filename: File name
    :param given_config: Config dictionary to update
    :return: Updated config dictionary
    """
    from copy import deepcopy

    from .constants import REQUIRED_CONFIGS
    from .constants import VALID_CONFIGS

    updated_config = deepcopy(given_config)

    try:
        file_config = raw_load_config(filename)
    except Exception as e:
        raise OSError(f"Error loading config from {filename}: {e}") from e

    # Check for unrecognized keys
    for key in file_config:
        if key not in VALID_CONFIGS:
            nice_list = "\n".join([f"  {item}" for item in VALID_CONFIGS])
            raise AttributeError(
                f"Invalid config parameter in {filename}: {key}.\nValid config parameters are:\n{nice_list}"
            )

    # Use default config as base and overwrite with given config keys as found
    for key in updated_config:
        if key in file_config:
            updated_config[key] = file_config[key]

    # Check for required keys
    for key in REQUIRED_CONFIGS:
        if key not in updated_config:
            raise AttributeError(f"{filename} is missing required '{key}'")

    # Validate and update sources
    updated_config = update_sources(updated_config)

    # Validate special cases
    validate_truth_configs(updated_config)
    validate_asr_configs(updated_config)

    # Check for non-empty spectral masks
    if len(updated_config["spectral_masks"]) == 0:
        updated_config["spectral_masks"] = given_config["spectral_masks"]

    return updated_config


def update_sources(given: dict) -> dict:
    """Validate and update fields in given 'sources'

    :param given: The dictionary of given config
    """
    from .constants import REQUIRED_NON_PRIMARY_SOURCE_CONFIGS
    from .constants import REQUIRED_SOURCE_CONFIGS
    from .constants import REQUIRED_SOURCES_CATEGORIES
    from .constants import VALID_NON_PRIMARY_SOURCE_CONFIGS
    from .constants import VALID_PRIMARY_SOURCE_CONFIGS

    sources = given["sources"]

    for category in REQUIRED_SOURCES_CATEGORIES:
        if category not in sources:
            raise AttributeError(f"config sources is missing required '{category}'")

    for category, source in sources.items():
        for key in REQUIRED_SOURCE_CONFIGS:
            if key not in source:
                raise AttributeError(f"config source '{category}' is missing required '{key}'")

        if category == "primary":
            for key in source:
                if key not in VALID_PRIMARY_SOURCE_CONFIGS:
                    nice_list = "\n".join([f"  {item}" for item in VALID_PRIMARY_SOURCE_CONFIGS])
                    raise AttributeError(
                        f"Invalid source '{category}' config parameter: '{key}'.\nValid sources config parameters are:\n{nice_list}"
                    )
        else:
            for key in REQUIRED_NON_PRIMARY_SOURCE_CONFIGS:
                if key not in source:
                    raise AttributeError(f"config source '{category}' is missing required '{key}'")

            for key in source:
                if key not in VALID_NON_PRIMARY_SOURCE_CONFIGS:
                    nice_list = "\n".join([f"  {item}" for item in VALID_NON_PRIMARY_SOURCE_CONFIGS])
                    raise AttributeError(
                        f"Invalid source '{category}' config parameter: '{key}'.\nValid source config parameters are:\n{nice_list}"
                    )

        files = source["files"]

        if isinstance(files, str) and files in sources and files != category:
            continue

        if isinstance(files, list):
            continue

        raise TypeError(
            f"'file' parameter of config source '{category}' is not a list or a reference to another source"
        )

    count = 0
    while any(isinstance(source["files"], str) for source in sources.values()) and count < 100:
        count += 1
        for category, source in sources.items():
            files = source["files"]
            if isinstance(files, str):
                given["sources"][category]["files"] = sources[files]["files"]

    if count == 100:
        raise RuntimeError("Check config sources for circular references")

    return given


def validate_truth_configs(given: dict) -> None:
    """Validate fields in given 'truth_configs'

    :param given: The dictionary of given config
    """
    from copy import deepcopy

    from . import truth_functions
    from .constants import REQUIRED_TRUTH_CONFIGS

    sources = given["sources"]

    for category, source in sources.items():
        if "truth_configs" not in source:
            continue

        truth_configs = source["truth_configs"]
        if len(truth_configs) == 0:
            raise ValueError(f"'truth_configs' in config source '{category}' is empty")

        for truth_name, truth_config in truth_configs.items():
            for k in REQUIRED_TRUTH_CONFIGS:
                if k not in truth_config:
                    raise AttributeError(
                        f"'{truth_name}' in source '{category}' truth_configs is missing required '{k}'"
                    )

            optional_config = deepcopy(truth_config)
            for k in REQUIRED_TRUTH_CONFIGS:
                del optional_config[k]

            getattr(truth_functions, truth_config["function"] + "_validate")(optional_config)


def validate_asr_configs(given: dict) -> None:
    """Validate fields in given 'asr_config'

    :param given: The dictionary of given config
    """
    from ..utils.asr import validate_asr
    from .constants import REQUIRED_ASR_CONFIGS

    if "asr_configs" not in given:
        raise AttributeError("config is missing required 'asr_configs'")

    asr_configs = given["asr_configs"]

    for name, asr_config in asr_configs.items():
        for key in REQUIRED_ASR_CONFIGS:
            if key not in asr_config:
                raise AttributeError(f"'{name}' in asr_configs is missing required '{key}'")

        engine = asr_config["engine"]
        config = {x: asr_config[x] for x in asr_config if x != "engine"}
        validate_asr(engine, **config)


def get_hierarchical_config_files(root: str, leaf: str) -> list[str]:
    """Get a hierarchical list of config files in the given leaf of the given root

    :param root: Root of the hierarchy
    :param leaf: Leaf under the root
    :return: List of config files found in the hierarchy
    """
    import os
    from pathlib import Path

    config_file = "config.yml"

    root_path = Path(os.path.abspath(root))
    if not root_path.is_dir():
        raise OSError(f"Given root, {root_path}, is not a directory.")

    leaf_path = Path(os.path.abspath(leaf))
    if not leaf_path.is_dir():
        raise OSError(f"Given leaf, {leaf_path}, is not a directory.")

    common = os.path.commonpath((root_path, leaf_path))
    if os.path.normpath(common) != os.path.normpath(root_path):
        raise OSError(f"Given leaf, {leaf_path}, is not in the hierarchy of the given root, {root_path}")

    top_config_file = os.path.join(root_path, config_file)
    if not Path(top_config_file).is_file():
        raise OSError(f"Could not find {top_config_file}")

    current = leaf_path
    config_files = []
    while current != root_path:
        local_config_file = Path(os.path.join(current, config_file))
        if local_config_file.is_file():
            config_files.append(str(local_config_file))
        current = current.parent

    config_files.append(top_config_file)
    return list(reversed(config_files))


def update_config_from_hierarchy(root: str, leaf: str, config: dict) -> dict:
    """Update the given config using the hierarchical config files in the given leaf of the given root

    :param root: Root of the hierarchy
    :param leaf: Leaf under the root
    :param config: Config to update
    :return: Updated config
    """
    from copy import deepcopy

    new_config = deepcopy(config)
    config_files = get_hierarchical_config_files(root=root, leaf=leaf)
    for config_file in config_files:
        new_config = update_config_from_file(filename=config_file, given_config=new_config)

    return new_config


def get_source_files(config: dict, show_progress: bool = False) -> list[SourceFile]:
    """Get the list of source files from a config

    :param config: Config dictionary
    :param show_progress: Show progress bar
    :return: List of source files
    """
    from itertools import chain

    from ..utils.parallel import par_track
    from ..utils.parallel import track

    sources = config["sources"]
    if not isinstance(sources, dict) and not all(isinstance(source, dict) for source in sources):
        raise TypeError("'sources' must be a dictionary of dictionaries")

    if "primary" not in sources:
        raise AttributeError("'primary' is missing in 'sources'")

    class_indices = config["class_indices"]
    if not isinstance(class_indices, list):
        class_indices = [class_indices]

    level_type = config["level_type"]

    source_files: list[SourceFile] = []
    for category in sources:
        source_files.extend(
            chain.from_iterable(
                [
                    append_source_files(
                        category=category,
                        entry=entry,
                        class_indices=class_indices,
                        truth_configs=sources[category].get("truth_configs", []),
                        level_type=level_type,
                    )
                    for entry in sources[category]["files"]
                ]
            )
        )

    progress = track(total=len(source_files), disable=not show_progress)
    source_files = par_track(_get_num_samples, source_files, progress=progress)
    progress.close()

    num_classes = config["num_classes"]
    for source_file in source_files:
        if any(class_index < 0 for class_index in source_file.class_indices):
            raise ValueError("class indices must contain only positive elements")

        if any(class_index > num_classes for class_index in source_file.class_indices):
            raise ValueError(f"class index elements must not be greater than {num_classes}")

    return source_files


def append_source_files(
    category: str,
    entry: dict,
    class_indices: list[int],
    truth_configs: dict,
    level_type: str,
    tokens: dict | None = None,
) -> list[SourceFile]:
    """Process source files list and append as needed

    :param category: Source file category name
    :param entry: Source file entry to append to the list
    :param class_indices: Class indices
    :param truth_configs: Truth configs
    :param level_type: Level type
    :param tokens: Tokens used for variable expansion
    :return: List of source files
    """
    from copy import deepcopy
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from ..datatypes import TruthConfig
    from ..utils.dataclass_from_dict import dataclass_from_dict
    from ..utils.tokenized_shell_vars import tokenized_expand
    from ..utils.tokenized_shell_vars import tokenized_replace
    from .audio import validate_input_file
    from .constants import REQUIRED_TRUTH_CONFIGS

    if tokens is None:
        tokens = {}

    truth_configs_merged = deepcopy(truth_configs)

    if not isinstance(entry, dict):
        raise TypeError("'entry' must be a dictionary")

    in_name = entry.get("name")
    if in_name is None:
        raise KeyError("Source file list contained record without name")

    class_indices = entry.get("class_indices", class_indices)
    if not isinstance(class_indices, list):
        class_indices = [class_indices]

    truth_configs_override = entry.get("truth_configs", {})
    for key in truth_configs_override:
        if key not in truth_configs:
            raise AttributeError(
                f"Truth config '{key}' override specified for {entry['name']} is not defined at top level"
            )
        if key in truth_configs_override:
            truth_configs_merged[key] |= truth_configs_override[key]

    level_type = entry.get("level_type", level_type)

    in_name, new_tokens = tokenized_expand(in_name)
    tokens.update(new_tokens)
    names = sorted(glob(in_name))
    if not names:
        raise OSError(f"Could not find {in_name}. Make sure path exists")

    source_files: list[SourceFile] = []
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                source_files.extend(
                    append_source_files(
                        category=category,
                        entry={"name": child},
                        class_indices=class_indices,
                        truth_configs=truth_configs_merged,
                        level_type=level_type,
                        tokens=tokens,
                    )
                )
        else:
            try:
                if ext == ".txt":
                    with open(file=name) as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition("#")[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = tokenized_expand(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                source_files.extend(
                                    append_source_files(
                                        category=category,
                                        entry={"name": child},
                                        class_indices=class_indices,
                                        truth_configs=truth_configs_merged,
                                        level_type=level_type,
                                        tokens=tokens,
                                    )
                                )
                else:
                    validate_input_file(name)
                    source_file = SourceFile(
                        category=category,
                        name=tokenized_replace(name, tokens),
                        samples=0,
                        class_indices=class_indices,
                        level_type=level_type,
                        truth_configs={},
                    )
                    if len(truth_configs_merged) > 0:
                        for tc_key, tc_value in truth_configs_merged.items():
                            config = deepcopy(tc_value)
                            truth_config: dict = {}
                            for key in REQUIRED_TRUTH_CONFIGS:
                                truth_config[key] = config[key]
                                del config[key]
                            truth_config["config"] = config
                            source_file.truth_configs[tc_key] = dataclass_from_dict(TruthConfig, truth_config)
                        for tc_key in source_file.truth_configs:
                            if (
                                "function" in truth_configs_merged[tc_key]
                                and truth_configs_merged[tc_key]["function"] == "file"
                            ):
                                truth_configs_merged[tc_key]["file"] = splitext(source_file.name)[0] + ".h5"
                    source_files.append(source_file)
            except Exception as e:
                raise OSError(f"Error processing {name}: {e}") from e

    return source_files


def get_ir_files(config: dict, show_progress: bool = False) -> list[ImpulseResponseFile]:
    """Get the list of impulse response files from a config

    :param config: Config dictionary
    :param show_progress: Show progress bar
    :return: List of impulse response files
    """
    from itertools import chain

    from ..utils.parallel import par_track
    from ..utils.parallel import track

    ir_files = list(
        chain.from_iterable(
            [
                append_ir_files(
                    entry=ImpulseResponseFile(
                        name=entry["name"],
                        tags=entry.get("tags", []),
                        delay=entry.get("delay", "auto"),
                    )
                )
                for entry in config["impulse_responses"]
            ]
        )
    )

    if len(ir_files) == 0:
        return []

    progress = track(total=len(ir_files), disable=not show_progress)
    ir_files = par_track(_get_ir_delay, ir_files, progress=progress)
    progress.close()

    return ir_files


def append_ir_files(entry: ImpulseResponseFile, tokens: dict | None = None) -> list[ImpulseResponseFile]:
    """Process impulse response files list and append as needed

    :param entry: Impulse response file entry to append to the list
    :param tokens: Tokens used for variable expansion
    :return: List of impulse response files
    """
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from ..utils.tokenized_shell_vars import tokenized_expand
    from ..utils.tokenized_shell_vars import tokenized_replace
    from .audio import validate_input_file

    if tokens is None:
        tokens = {}

    in_name, new_tokens = tokenized_expand(entry.name)
    tokens.update(new_tokens)
    names = sorted(glob(in_name))
    if not names:
        raise OSError(f"Could not find {in_name}. Make sure path exists")

    ir_files: list[ImpulseResponseFile] = []
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                if not isabs(file):
                    file = join(dir_name, file)
                child = ImpulseResponseFile(file, entry.tags, entry.delay)
                ir_files.extend(append_ir_files(entry=child, tokens=tokens))
        else:
            try:
                if ext == ".txt":
                    with open(file=name) as txt_file:
                        for line in txt_file:
                            # strip comments
                            file = line.partition("#")[0]
                            file = file.rstrip()
                            if file:
                                file, new_tokens = tokenized_expand(file)
                                tokens.update(new_tokens)
                                if not isabs(file):
                                    file = join(dir_name, file)
                                child = ImpulseResponseFile(file, entry.tags, entry.delay)
                                ir_files.extend(append_ir_files(entry=child, tokens=tokens))
                elif ext == ".yml":
                    try:
                        yml_config = raw_load_config(name)

                        if "impulse_responses" in yml_config:
                            for record in yml_config["impulse_responses"]:
                                ir_files.extend(append_ir_files(entry=record, tokens=tokens))
                    except Exception as e:
                        raise OSError(f"Error processing {name}: {e}") from e
                else:
                    validate_input_file(name)
                    ir_files.append(ImpulseResponseFile(tokenized_replace(name, tokens), entry.tags, entry.delay))
            except Exception as e:
                raise OSError(f"Error processing {name}: {e}") from e

    return ir_files


def get_spectral_masks(config: dict) -> list[SpectralMask]:
    """Get the list of spectral masks from a config

    :param config: Config dictionary
    :return: List of spectral masks
    """
    from ..utils.dataclass_from_dict import list_dataclass_from_dict

    try:
        return list_dataclass_from_dict(list[SpectralMask], config["spectral_masks"])
    except Exception as e:
        raise ValueError(f"Error in spectral_masks: {e}") from e


def get_truth_parameters(config: dict) -> list[TruthParameter]:
    """Get the list of truth parameters from a config

    :param config: Config dictionary
    :return: List of truth parameters
    """
    from copy import deepcopy

    from . import truth_functions
    from .constants import REQUIRED_TRUTH_CONFIGS

    truth_parameters: list[TruthParameter] = []
    for category, source_config in config["sources"].items():
        if "truth_configs" in source_config:
            for truth_name, truth_config in source_config["truth_configs"].items():
                optional_config = deepcopy(truth_config)
                for key in REQUIRED_TRUTH_CONFIGS:
                    del optional_config[key]

                parameters = getattr(truth_functions, truth_config["function"] + "_parameters")(
                    config["feature"],
                    config["num_classes"],
                    optional_config,
                )
                truth_parameters.append(TruthParameter(category, truth_name, parameters))

    return truth_parameters


def _get_num_samples(entry: SourceFile) -> SourceFile:
    from .audio import get_num_samples

    entry.samples = get_num_samples(entry.name)
    return entry


def _get_ir_delay(entry: ImpulseResponseFile) -> ImpulseResponseFile:
    from .ir_delay import get_ir_delay

    if entry.delay == "auto":
        entry.delay = get_ir_delay(entry.name)
    else:
        try:
            entry.delay = int(entry.delay)
        except ValueError as e:
            raise ValueError(f"Invalid impulse response delay: {entry.delay}") from e

    return entry



import contextlib


@contextlib.contextmanager
def seed_context(seed):
    import numpy as np

    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)



from collections.abc import Sequence

from onnx import ModelProto
from onnx import ValueInfoProto
from onnxruntime import InferenceSession
from onnxruntime import NodeArg  # pyright: ignore [reportAttributeAccessIssue]
from onnxruntime import SessionOptions  # pyright: ignore [reportAttributeAccessIssue]

REQUIRED_HPARAMS = ("feature", "batch_size", "timesteps")


def _extract_shapes(io: list[ValueInfoProto]) -> list[list[int] | str]:
    shapes: list[list[int] | str] = []

    # iterate through inputs of the graph to find shapes
    for item in io:
        # get tensor type: 0, 1, 2, etc.
        tensor_type = item.type.tensor_type
        # check if it has a shape
        if tensor_type.HasField("shape"):
            tmp_shape = []
            # iterate through dimensions of the shape
            for d in tensor_type.shape.dim:
                if d.HasField("dim_value"):
                    # known dimension, int value
                    tmp_shape.append(d.dim_value)
                elif d.HasField("dim_param"):
                    # dynamic dim with symbolic name of d.dim_param; set size to 0
                    tmp_shape.append(0)
                else:
                    # unknown dimension with no name; also set to 0
                    tmp_shape.append(0)
            # add as a list
            shapes.append(tmp_shape)
        else:
            shapes.append("unknown rank")

    return shapes


def get_and_check_inputs(model: ModelProto) -> tuple[list[ValueInfoProto], list[list[int] | str]]:
    from sonusai import logger

    # ignore initializer inputs (only seen in older ONNX < v1.5)
    initializer_names = [x.name for x in model.graph.initializer]
    inputs = [i for i in model.graph.input if i.name not in initializer_names]
    if len(inputs) != 1:
        logger.warning(f"Warning: ONNX model has {len(inputs)} inputs; expected only 1")

    # This one-liner works only if input has type and shape, returns a list
    # shape0 = [d.dim_value for d in inputs[0].type.tensor_type.shape.dim]
    shapes = _extract_shapes(inputs)

    return inputs, shapes


def get_and_check_outputs(model: ModelProto) -> tuple[list[ValueInfoProto], list[list[int] | str]]:
    from sonusai import logger

    outputs = list(model.graph.output)
    if len(outputs) != 1:
        logger.warning(f"Warning: ONNX model has {len(outputs)} outputs; expected only 1")

    shapes = _extract_shapes(outputs)

    return outputs, shapes


def add_sonusai_metadata(model: ModelProto, hparams: dict) -> ModelProto:
    """Add SonusAI hyperparameters as metadata to an ONNX model using 'hparams' key

    :param model: ONNX model
    :param hparams: dictionary of hyperparameters to add
    :return: ONNX model

    Note SonusAI conventions require models to have:
        feature: Model feature type
        batch_size: Model batch size
        timesteps: Size of timestep dimension (0 for no dimension)
    """
    from sonusai import logger

    # Note hparams should be a dict (i.e., extracted from checkpoint)
    if eval(str(hparams)) != hparams:  # noqa: S307
        raise TypeError("hparams is not a dict")
    for key in REQUIRED_HPARAMS:
        if key not in hparams:
            logger.warning(f"Warning: SonusAI hyperparameters are missing: {key}")

    meta = model.metadata_props.add()
    meta.key = "hparams"
    meta.value = str(hparams)

    return model


def get_sonusai_metadata(session: InferenceSession) -> dict | None:
    """Get SonusAI hyperparameter metadata from an ONNX Runtime session."""
    from sonusai import logger

    meta = session.get_modelmeta()
    if "hparams" not in meta.custom_metadata_map:
        logger.warning("Warning: ONNX model metadata does not contain 'hparams'")
        return None

    hparams = eval(meta.custom_metadata_map["hparams"])  # noqa: S307
    for key in REQUIRED_HPARAMS:
        if key not in hparams:
            logger.warning(f"Warning: ONNX model does not have required SonusAI hyperparameters: {key}")

    return hparams


def load_ort_session(
    model_path: str, providers: Sequence[str | tuple[str, dict]] | None = None
) -> tuple[InferenceSession, SessionOptions, str, dict | None, list[NodeArg], list[NodeArg]]:
    from os.path import basename
    from os.path import exists
    from os.path import isfile
    from os.path import splitext

    import onnxruntime as ort

    from sonusai import logger

    if providers is None:
        providers = ["CPUExecutionProvider"]

    if exists(model_path) and isfile(model_path):
        model_basename = basename(model_path)
        model_root = splitext(model_basename)[0]
        logger.info(f"Importing model from {model_basename}")
        try:
            session = ort.InferenceSession(model_path, providers=providers)
            options = ort.SessionOptions()
        except Exception as e:
            logger.exception(f"Error: could not load ONNX model from {model_path}: {e}")
            raise SystemExit(1) from e
    else:
        logger.exception(f"Error: model file does not exist: {model_path}")
        raise SystemExit(1)

    logger.info(f"Opened session with provider options: {session._provider_options}.")
    hparams = get_sonusai_metadata(session)
    if hparams is not None:
        for key in REQUIRED_HPARAMS:
            logger.info(f"  {key:12} {hparams[key]}")

    inputs = session.get_inputs()
    outputs = session.get_outputs()

    # in_names = [n.name for n in session.get_inputs()]
    # out_names = [n.name for n in session.get_outputs()]

    return session, options, model_root, hparams, inputs, outputs



# SonusAI query utilities
# ruff: noqa: F401

from .queries import get_mixids_from_class_indices
from .queries import get_mixids_from_noise
from .queries import get_mixids_from_snr
from .queries import get_mixids_from_source
from .queries import get_mixids_from_truth_function



from ..datatypes import AudioT


def resample(audio: AudioT, orig_sr: int, target_sr: int) -> AudioT:
    from librosa import resample

    return resample(audio, orig_sr=orig_sr, target_sr=target_sr, res_type="soxr_hq")



"""sonusai onnx_predict

usage: onnx_predict [-hvlwr] [--include GLOB] [-i MIXID] MODEL DATA ...

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to generate if input is a mixture database. [default: *].
    --include GLOB              Search only files whose base name matches GLOB. [default: *.{wav,flac}].
    -w, --write-wav             Calculate inverse transform of prediction and write .wav files

Run prediction (inference) using an ONNX model on a SonusAI mixture dataset or audio files from a glob path.
The ONNX Runtime (ort) inference engine is used to execute the inference.

Inputs:
    MODEL       ONNX model .onnx file of a trained model (weights are expected to be in the file).
                The model must also include required Sonusai hyperparameters.  See theSonusai torchl_onnx command.

    DATA        A string which must be one of the following:
                1. Path to a single file.  The prediction data is written to <filename_predict.*> in same location.
                2. Path to a Sonusai Mixture Database directory.
                 - Sonusai mixture database directory, prediction files will be named mixid_predict.*
                 - MIXID will select a subset of mixture ids
                3. Directory with audio files found recursively within.  See GLOB audio file extensions below.
                4. Regex resolving to a list of files.
                 - Subdirectory containing audio files with extension
                 - Regex resolving to a list of audio files

                   generate feature and truth data if not found.

Note there are multiple ways to process model prediction over multiple audio data files:
1. TSE (timestep single extension): mixture transform frames are fit into the timestep dimension and the model run as
   a single inference call.  If batch_size is > 1 then run multiple mixtures in one call with shorter mixtures
   zero-padded to the size of the largest mixture.
2. TME (timestep multi-extension): mixture is split into multiple timesteps, i.e. batch[0] is starting timesteps, ...
   Note that batches are run independently, thus sequential state from one set of timesteps to the next will not be
   maintained, thus results for such models (i.e. conv, LSTMs, in the timestep dimension) would not match using
   TSE mode.

TBD not sure below make sense, need to continue ??
2. BSE (batch single extension): mixture transform frames are fit into the batch dimension. This make sense only if
   independent predictions are made on each frame w/o considering previous frames (timesteps=1) or there is no
   timestep dimension in the model (timesteps=0).
3. Classification

Outputs the following to opredict-<TIMESTAMP> directory:
    <id>
        predict.h5
    onnx_predict.log

"""


def process_path(path, ext_list: list[str] | None = None):
    """
    Check path which can be a single file, a subdirectory, or a regex
    return:
      - a list of files with matching extensions to any in ext_list provided (i.e. ['.wav', '.mp3', '.acc'])
      - the basedir of the path, if
    """
    import glob
    from os.path import abspath
    from os.path import commonprefix
    from os.path import dirname
    from os.path import isdir
    from os.path import isfile
    from os.path import join

    from sonusai.utils import braced_iglob

    if ext_list is None:
        ext_list = [".wav", ".WAV", ".flac", ".FLAC", ".mp3", ".aac"]

    # Check if the path is a single file, and return it as a list with the dirname
    if isfile(path):
        if any(path.endswith(ext) for ext in ext_list):
            basedir = dirname(path)  # base directory
            if not basedir:
                basedir = "./"
            return [path], basedir
        else:
            return [], []

    # Check if the path is a dir, recursively find all files any of the specified extensions, return file list and dir
    if isdir(path):
        matching_files = []
        for ext in ext_list:
            matching_files.extend(glob.glob(join(path, "**/*" + ext), recursive=True))
        return matching_files, path

    # Process as a regex, return list of filenames and basedir
    apath = abspath(path)  # join(abspath(path), "**", "*.{wav,flac,WAV}")
    matching_files = []
    for file in braced_iglob(pathname=apath, recursive=True):
        matching_files.append(file)
    if matching_files:
        basedir = commonprefix(matching_files)  # Find basedir
        return matching_files, basedir
    else:
        return [], []


def main() -> None:
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    verbose = args["--verbose"]
    wav = args["--write-wav"]
    mixids = args["--mixid"]
    include = args["--include"]
    model_path = args["MODEL"]
    data_paths = args["DATA"]

    # Quick check of CPU and GPU devices
    import re
    import subprocess
    import time
    from os import makedirs
    from os.path import basename
    from os.path import exists
    from os.path import isdir
    from os.path import isfile
    from os.path import join
    from os.path import normpath
    from os.path import splitext

    import h5py
    import numpy as np
    import onnxruntime as ort
    import psutil

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import get_audio_from_feature
    from sonusai.utils import create_ts_name
    from sonusai.utils import load_ort_session
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import write_audio

    num_cpu = psutil.cpu_count()
    cpu_percent = psutil.cpu_percent(interval=1)
    print(f"#CPUs: {num_cpu}, current CPU utilization: {cpu_percent}%")
    print(f"Memory utilization: {psutil.virtual_memory().percent}%")

    vga_devices = [
        line.split(" ", 3)[-1]
        for line in subprocess.check_output("lspci | grep -i vga", shell=True).decode().splitlines()
    ]
    nv_devs = list(filter(lambda x: "nvidia" in x.lower(), vga_devices))
    nv_mods = [re.search(r"\[.*?\]", device).group(0) if re.search(r"\[.*?\]", device) else None for device in nv_devs]
    if len(nv_mods) > 0:
        print(f"{len(nv_mods)} Nvidia devices present: {nv_mods}")  # prints model names
    else:
        print("No cuda devices present, using cpu")

    avail_providers = ort.get_available_providers()
    print(f"Loaded ONNX Runtime, available providers: {avail_providers}.")
    if len(nv_mods) > 0:
        print(
            "If GPU is desired, need to replace onnxruntime with onnxruntime-gpu i.e. using pip:"
            "> pip uninstall onnxruntime"
            "> pip install onnxruntime-gpu\n\n"
        )

    # Quick check that model is valid
    if exists(model_path) and isfile(model_path):
        try:
            session = ort.InferenceSession(model_path)
            options = ort.SessionOptions()
        except Exception as e:
            print(f"Error: could not load ONNX model from {model_path}: {e}")
            raise SystemExit(1) from e
    else:
        print(f"Error: model file path is not valid: {model_path}")
        raise SystemExit(1)

    # Check datapath is valid
    if len(data_paths) == 1 and isdir(data_paths[0]):  # Try opening as mixdb subdir
        mixdb_path = data_paths[0]
        try:
            mixdb = MixtureDatabase(mixdb_path)
        except Exception:
            mixdb_path = None
        in_basename = basename(normpath(data_paths[0]))
        output_dir = create_ts_name("opredict-" + in_basename)
        num_featparams = mixdb.feature_parameters
        print(f"Loaded SonusAI mixdb with {mixdb.num_mixtures} mixtures and {num_featparams} classes")
        p_mixids = mixdb.mixids_to_list(mixids)
        feature_mode = mixdb.feature

    if mixdb_path is None:
        if verbose:
            print(f"Checking {len(data_paths)} locations ... ")
        # Check location, default ext are ['.wav', '.WAV', '.flac', '.FLAC', '.mp3', '.aac']
        pfiles, basedir = process_path(data_paths)
        if pfiles is None or len(pfiles) < 1:
            print(f"No audio files or Sonusai mixture database found in {data_paths}, exiting ...")
            raise SystemExit(1)
        else:
            pfiles = sorted(pfiles, key=basename)
            output_dir = basedir

    if mixdb_path is not None or len(pfiles) > 1:  # log file only if mixdb or more than one file
        makedirs(output_dir, exist_ok=True)
        # Setup logging file
        create_file_handler(join(output_dir, "onnx-predict.log"))
        update_console_handler(verbose)
        initial_log_messages("onnx_predict")
        # print some previous messages
        logger.info(f"Loaded ONNX Runtime, available providers: {avail_providers}.")
        if mixdb_path:
            logger.debug(f"Loaded SonusAI mixdb with {mixdb.num_mixtures} mixtures and {num_featparams} classes")
            if len(p_mixids) != mixdb.num_mixtures:
                logger.info(f"Processing a subset of {len(p_mixids)} from available mixtures.")

    # Reload model/session and do more thorough checking
    session, options, model_root, hparams, sess_inputs, sess_outputs = load_ort_session(model_path)
    if "CUDAExecutionProvider" in avail_providers:
        session.set_providers(["CUDAExecutionProvider"])
    if hparams is None:
        logger.error("Error: ONNX model does not have required SonusAI hyperparameters, cannot proceed.")
        raise SystemExit(1)

    if len(sess_inputs) != 1:  # TBD update to support state_in and state_out
        logger.error(f"Error: ONNX model does not have 1 input, but {len(sess_inputs)}. Exit due to unknown input.")

    in0name = sess_inputs[0].name
    in0type = sess_inputs[0].type
    in0shape = sess_inputs[0].shape  # a list
    # Check for 2 cases of model feature input shape:  batch x timesteps x fparams or batch x fparams
    if not isinstance(in0shape[0], str):
        model_batchsz = int(in0shape[0])
        logger.debug(f"Onnx model has fixed batch_size: {model_batchsz}.")
    else:
        model_batchsz = -1
        logger.debug("Onnx model has a dynamic batch_size.")

    if len(in0shape) < 3:
        model_tsteps = 0
        model_featparams = int(in0shape[1])
    else:
        model_featparams = int(in0shape[2])
        if not isinstance(in0shape[1], str):
            model_tsteps = int(in0shape[1])
            logger.debug(f"Onnx model has fixed timesteps: {model_tsteps}.")
        else:
            model_tsteps = -1
            logger.debug("Onnx model has dynamic timesteps dimension size.")

    out_names = [n.name for n in session.get_outputs()]

    if in0type.find("float16") != -1:
        model_is_fp16 = True
        logger.info("Detected input of float16, converting all feature inputs to that type.")
    else:
        model_is_fp16 = False

    logger.info(f"Read and compiled ONNX model from {model_path}.")

    start_time = time.monotonic()

    if mixdb is not None and hparams["batch_size"] == 1:
        if hparams["feature"] != feature_mode:  # warn on mis-match, but TBD could be sov-mode
            logger.warning("Mixture feature does not match model feature, this inference run may fail.")
        logger.info(f"Processing {len(p_mixids)} mixtures from SonusAI mixdb ...")
        logger.info(f"Using OnnxRT provider {session.get_providers()} ...")

        for mixid in p_mixids:
            # feature data is now always fp32 and frames x stride x feature_params
            feat_dat, _ = mixdb.mixture_ft(mixid)
            if feat_dat.shape[1] > 1:  # stride mode num_frames overrides batch dim, no reshape
                stride_mode = 1
                batch_size = feat_dat.shape[0]  # num_frames in stride mode becomes batch size
            if hparams["timesteps"] == 0:
                # no timestep dimension, remove the dimension
                timesteps = 0
                feat_dat = np.reshape(feat_dat, [batch_size, num_featparams])
            else:
                # fit frames into timestep dimension (TSE mode) and knowing batch_size = 1
                timesteps = feat_dat.shape[0]
                feat_dat = np.transpose(feat_dat, (1, 0, 2))  # transpose to 1 x frames=tsteps x feat_params

            if model_is_fp16:
                feat_dat = np.float16(feat_dat)  # type: ignore[assignment]

            # run inference, ort session wants i.e. batch x timesteps x feat_params, outputs numpy BxTxFP or BxFP
            predict = session.run(out_names, {in0name: feat_dat})[0]
            # predict, _ = reshape_outputs(predict=predict[0], timesteps=frames)  # frames x feat_params

            output_fname = join(output_dir, mixdb.mixture(mixid).name)
            with h5py.File(output_fname + ".h5", "a") as f:
                if "predict" in f:
                    del f["predict"]
                f.create_dataset("predict", data=predict)
            if wav:
                # note only makes sense if model is predicting audio, i.e., timestep dimension exists
                # predict_audio wants [frames, channels, feature_parameters] equivalent to timesteps, batch, bins
                predict = np.transpose(predict, [1, 0, 2])
                predict_audio = get_audio_from_feature(feature=predict, feature_mode=feature_mode)
                owav_name = splitext(output_fname)[0] + "_predict.wav"
                write_audio(owav_name, predict_audio)

    else:  # TBD add support
        logger.info("Mixture database does not exist or batch_size is not equal to one, exiting ...")

    end_time = time.monotonic()
    logger.info(f"Completed in {seconds_to_hms(seconds=end_time - start_time)}")
    logger.info("")


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)

    # mixdb_path = None
    # mixdb: MixtureDatabase | None = None
    # p_mixids: list[int] = []
    # entries: list[PathInfo] = []
    #
    # if len(data_paths) == 1 and isdir(data_paths[0]):
    #     # Assume it's a single path to SonusAI mixdb subdir
    #     in_basename = basename(normpath(data_paths[0]))
    #     mixdb_path = data_paths[0]
    # else:
    #     # search all data paths for .wav, .flac (or whatever is specified in include)
    #     in_basename = ""

    # if mixdb_path is not None:  # a mixdb is found and loaded
    #     # Assume it's a single path to SonusAI mixdb subdir
    #     num_featparams = mixdb.feature_parameters
    #     logger.debug(f"SonusAI mixdb: found {mixdb.num_mixtures} mixtures with {num_featparams} classes")
    #     p_mixids = mixdb.mixids_to_list(mixids)
    #     if len(p_mixids) != mixdb.num_mixtures:
    #         logger.info(f"Processing a subset of {p_mixids} from available mixtures.")
    # else:
    # for p in data_paths:
    #     location = join(realpath(abspath(p)), "**", include)
    #     logger.debug(f"Processing files in {location}")
    #     for file in braced_iglob(pathname=location, recursive=True):
    #         name = file
    #         entries.append(PathInfo(abs_path=file, audio_filepath=name))
    # logger.info(f"{len(data_paths)} data paths specified, found {len(pfile)} audio files.")

    # feature, _ = reshape_inputs(
    #     feature=feature,
    #     batch_size=1,
    #     timesteps=timesteps,
    #     flatten=hparams["flatten"],
    #     add1ch=hparams["add1ch"],
    # )



from ..mixture.config import get_default_config


def doc_seed() -> str:
    default = f"\nDefault value: {get_default_config()['seed']}"
    # fmt: off
    return """
'seed' is a mixture database configuration parameter that sets the random number
generator seed.
""" + default
    # fmt: on


def doc_feature() -> str:
    default = f"\nDefault value: {get_default_config()['feature']}"
    # fmt: off
    return """
'feature' is a mixture database configuration parameter that sets the feature
to use.
""" + default
    # fmt: on


def doc_target_level_type() -> str:
    default = f"\nDefault value: {get_default_config()['target_level_type']}"
    # fmt: off
    return """
'target_level_type' is a mixture database configuration parameter that sets the
algorithm to use to determine target energy level for SNR calculations.
Supported values are:

 default    mean of squares
 speech     ITU-T P.56 active speech level method B
""" + default
    # fmt: on


def doc_targets() -> str:
    default = f"\nDefault value: {get_default_config()['targets']}"
    # fmt: off
    return """
'targets' is a mixture database configuration parameter that sets the list of
targets to use.

Required field:

 'name'
                File name. May be one of the following:

  audio         Supported formats are .wav, .mp3, .m4a, .aif, .flac, and .ogg
  glob          Matches file glob patterns
  .yml          The given YAML file is parsed into the list
  .txt          Each line in the given text file indicates an item which
                may be anything in this list (audio, glob, .yml, or .txt)

Optional fields:

 'truth_configs'
                Local overrides for truth configs. Contains the following:
  'name'        Name of truth config
    '<param1>'  Target-specific override for truth configuration parameter
    '<param2>'  Target-specific override for truth configuration parameter

  'target_level_type'
                Target-specific override for target_level_type.

Example:

targets:
  - name: data/esc50/ESC-50-master/audio/1-*.wav
    truth_configs:
      sed:
        thresholds: [-38, -41, -48]
        index: 2
        class_balancing_augmentation: { }
  - name: target.mp3
    truth_configs:
      sed:
        thresholds: [-37, -40, -46]
        index: 5
""" + default
    # fmt: on


def doc_num_classes() -> str:
    default = f"\nDefault value: {get_default_config()['num_classes']}"
    # fmt: off
    return """
'num_classes' is a mixture database configuration parameter that sets the number of
classes in this dataset. The number of classes is the total number of parameters
(or classes or labels) in the truth. This controls the size of the truth input to
the model.

Note that the model output 'parameters' dimension is NOT necessarily the same size
as the truth 'num_classes' dimension; there may be multiple truth functions combined
in the truth, e.g., for use in loss function calculations.
""" + default
    # fmt: on


def doc_class_labels() -> str:
    default = f"\nDefault value: {get_default_config()['class_labels']}"
    # fmt: off
    return """
'class_labels' is a mixture database configuration parameter that sets class labels
in this dataset.
""" + default
    # fmt: on


# def doc_class_weights_threshold() -> str:
#     default = f"\nDefault value: {get_default_config()['class_weights_threshold']}"
#     # fmt: off
#     return """
# 'class_weights_threshold' is a mixture database configuration parameter that sets
# the threshold for class weights calculation to quantize truth to binary for counting.
#
# Supports scalar or list:
#
#  scalar     use for all classes
#  list       must be of num_classes length
# """ + default
#     # fmt: on


def get_truth_functions() -> str:
    from ..mixture import truth_functions

    functions = [function for function in dir(truth_functions) if not function.startswith("__")]
    text = "\nSupported truth functions:\n\n"
    for function in functions:
        docs = getattr(truth_functions, function).__doc__
        if docs is not None:
            text += f"    {function}\n"
            for doc in docs.splitlines():
                text += f"        {doc}\n"
    return text


def doc_truth_configs() -> str:
    import yaml

    default = f"\nDefault value:\n\n{yaml.dump(get_default_config()['truth_configs'])}"
    # fmt: off
    return """
'truth_configs' is a mixture database configuration parameter that sets the truth
generation configurations for targets. There is a global 'truth_configs' and there may be
target-specific 'truth_configs'.

A truth config creates a type of truth and is associated with target file(s).
Target files may have multiple truth settings.

Truth is always generated per transform frame.

Note that there is a difference between transform frames and feature frames: a feature
frame may be decimated and may have a stride dimension greater than 1 (which aggregates
multiple transform frames in a single feature).

There are two notions of truth data: truth_t and truth_f. truth_t is what the truth
functions always generate and is in the transform frame domain [transform_frames, truth_parameters].
truth_f, or truth in the feature domain, is created by passing truth_t into the feature
generator which produces feature frame domain truth data [feature_frames, stride or 1, truth_parameters].

The stride dimension may be reduced using the 'stride_reduction' parameter. Supported stride
reduction methods:
    'none'      preserve the stride dimension with no change
    'max'       reduce the stride dimension to 1 by taking the max
    'mean'      reduce the stride dimension to 1 by taking the mean
    'first'     reduce the stride dimension to 1 by taking the value in the first stride index

The 'truth_configs' parameter specifies the following:

  'name'        Name of truth configuration
    'function'  Name of truth function to use
    'stride_reduction'
                Name of stride reduction method to use
    '<param1>'  Function-specific configuration parameter
    '<paramN>'  Function-specific configuration parameter
    'class_balancing_augmentation'
                Class balancing augmentation.
                This truth configuration will use this rule for class balancing operations.
                If this rule is empty or unspecified, then this truth function will not
                perform class balancing.

                Class balancing ensures that each class in a sound classification dataset
                is represented equally (i.e., each class has the same number of augmented
                targets). This is achieved by creating new class balancing augmentation
                rules and applying them to targets in underrepresented classes to create
                more augmented targets for those classes.

                This rule must contain at least one random entry in order to guarantee
                unique additional data.

                See 'augmentations' for details on augmentation rules.
""" + get_truth_functions() + default
    # fmt: on


def doc_augmentations() -> str:
    # fmt: off
    return """
Augmentation Rules

These rules may be specified for target and/or noise. Each rule will be
applied for each target/noise. The values may be specified as scalars, lists,
or random using the syntax: 'rand(<min>, <max>)'.

If a value is specified as a list, then the rule is repeated for each value in
the list.

If a value is specified using rand, then a randomized rule is generated
dynamically per use.

Rules may specify any or all of the following augmentations:

 'normalize'    Normalize audio file to the specified level (in dBFS).
 'gain'         Apply an amplification or an attenuation to the audio signal.
                The signal level is adjusted by the given number of dB; positive
                amplifies, negative attenuates, 0 does nothing.
 'pitch'        Change the audio pitch (but not its tempo). Pitch amount is
                specified as positive or negative 'cents' (i.e., 100ths of a
                semitone).
 'tempo'        Change the audio tempo (but not its pitch). Tempo amount is
                specified as the ratio of the new tempo to the old tempo. For
                example, '1.1' speeds up the tempo by 10% and '0.9' slows it
                down by 10%.
 'eq1'          Apply a two-pole peaking equalization filter. EQ parameters are
                specified as a [frequency, width, gain] triple where:
                  'frequency' gives the central frequency in Hz (20 - SR/2),
                  'width' gives the width as a Q-factor (0.3 - 2.0), and
                  'gain' gives the gain in dB (-20 - 20).
 'eq2'          Apply an additional band of EQ. Same as 'eq1'
 'eq3'          Apply an additional band of EQ. Same as 'eq1'
 'lpf'          Apply a low-pass Butterworth filter. The 3 dB point frequency is
                specified in Hz (20 - SR/2).
 'ir'           An index into a list of impulse responses (specified in the
                'impulse_responses' parameter).
                For targets, the impulse response is applied AFTER truth generation
                and the resulting audio is still aligned with the truth. Random
                syntax for 'ir' is simply 'rand' (i.e., do not specify <min> and <max>).

Only the specified augmentations for a given rule are applied; all others are
skipped in the given rule. For example, if a rule only specifies 'tempo',
then only a tempo augmentation is applied and all other possible augmentations
are ignored (e.g., 'gain', 'pitch', etc.).

Example:

target_augmentations:
  - normalize: -3.5
  - normalize: -3.5
    pitch: [-300, 300]
    tempo: [0.8, 1.2]
    eq1: [[1000, 0.8, 3], [600, 1.0, -4], [800, 0.6, 0]]
  - normalize: -3.5
    pitch: "rand(-300, 300)"
    eq1: ["rand(100, 6000)", "rand(0.6, 1.0)", "rand(-6, 6)"]
    lpf: "rand(1000, 8000)"
  - tempo: "rand(0.9, 1.1)"
    eq1: [["rand(100, 7500)", 0.8, -10], ["rand(100, 7500)", 0.8, 10]]

There are four rules given in this example.

The first rule is simple:
  - normalize: -3.5

This results in just one augmentation being applied to each target:

  normalize: -3.5

The second rule illustrates the use of lists to specify values:
  - normalize: -3.5
    pitch: [-300, 300]
    tempo: [0.8, 1.2]
    eq1: [[1000, 0.8, 3], [600, 1.0, -4], [800, 0.6, 0]]

There are two values given for pitch, two for tempo, and three for EQ. This
rule expands to 2 * 2 * 3 = 12 unique augmentations being applied to each
target:

  normalize: -3.5, pitch: -3, tempo: 0.8, eq1: [1000, 0.8,  3]
  normalize: -3.5, pitch: -3, tempo: 0.8, eq1: [ 600, 1.0, -4]
  normalize: -3.5, pitch: -3, tempo: 0.8, eq1: [ 800, 0.6,  0]
  normalize: -3.5, pitch: -3, tempo: 1.2, eq1: [1000, 0.8,  3]
  normalize: -3.5, pitch: -3, tempo: 1.2, eq1: [ 600, 1.0, -4]
  normalize: -3.5, pitch: -3, tempo: 1.2, eq1: [ 800, 0.6,  0]
  normalize: -3.5, pitch:  3, tempo: 0.8, eq1: [1000, 0.8,  3]
  normalize: -3.5, pitch:  3, tempo: 0.8, eq1: [ 600, 1.0, -4]
  normalize: -3.5, pitch:  3, tempo: 0.8, eq1: [ 800, 0.6,  0]
  normalize: -3.5, pitch:  3, tempo: 1.2, eq1: [1000, 0.8,  3]
  normalize: -3.5, pitch:  3, tempo: 1.2, eq1: [ 600, 1.0, -4]
  normalize: -3.5, pitch:  3, tempo: 1.2, eq1: [ 800, 0.6,  0]

The third rule shows the use of rand:
  - normalize: -3.5
    pitch: "rand(-300, 300)"
    eq1: ["rand(100, 6000)", "rand(0.6, 1.0)", "rand(-6, 6)"]
    lpf: "rand(1000, 8000)"

This rule is used to create randomized augmentations per use.

The fourth rule demonstrates the use of scalars, lists, and rand:
  - tempo: [0.9, 1, 1.1]
    eq1: [["rand(100, 7500)", 0.8, -10], ["rand(100, 7500)", 0.8, 10]]

This rule expands to 6 unique augmentations being applied to each target
(list of 3 * list of 2). Here is the expansion:

  tempo: 0.9, eq1: ["rand(100, 7500)", 0.8, -10]
  tempo: 1.0, eq1: ["rand(100, 7500)", 0.8, -10]
  tempo: 1.1, eq1: ["rand(100, 7500)", 0.8, -10]
  tempo: 0.9, eq1: ["rand(100, 7500)", 0.8, 10]
  tempo: 1.0, eq1: ["rand(100, 7500)", 0.8, 10]
  tempo: 1.1, eq1: ["rand(100, 7500)", 0.8, 10]"""
    # fmt: on


def doc_target_augmentations() -> str:
    import yaml

    default = f"\nDefault value:\n\n{yaml.dump(get_default_config()['target_augmentations'])}"
    # fmt: off
    return """
'target_augmentations' is a mixture database configuration parameter that
specifies a list of augmentation rules to use for each target.

See 'augmentations' for details on augmentation rules.
""" + default
    # fmt: on


def doc_target_distortions() -> str:
    import yaml

    default = f"\nDefault value:\n\n{yaml.dump(get_default_config()['target_distortions'])}"
    # fmt: off
    return """
'target_distortions' is a mixture database configuration parameter that
specifies a list of distortion rules to use for each target.

See 'augmentations' for details on distortion rules.
""" + default
    # fmt: on


def doc_noises() -> str:
    default = f"\nDefault value: {get_default_config()['class_balancing']}"
    # fmt: off
    return """
'noises' is a mixture database configuration parameter that sets the list of
noises to use.

Required field:

 'name'
                File name. May be one of the following:

   audio        Supported formats are .wav, .mp3, .aif, .flac, and .ogg
   glob         Matches file glob patterns
   .yml         The given YAML file is parsed into the list
   .txt         Each line in the given text file indicates an item which
                may be anything in this list (audio, glob, .yml, or .txt)
""" + default
    # fmt: on


def doc_noise_augmentations() -> str:
    import yaml

    default = f"\nDefault value:\n\n{yaml.dump(get_default_config()['noise_augmentations'])}"

    # fmt: off
    return """
'noise_augmentations' is a mixture database configuration parameter that
specifies a list of augmentation rules to use for each noise.

See 'augmentations' for details on augmentation rules.
""" + default
    # fmt: on


def doc_snrs() -> str:
    default = f"\nDefault value: {get_default_config()['snrs']}"
    # fmt: off
    return """
'snrs' is a mixture database configuration parameter that specifies a list
of required signal-to-noise ratios (in dB).

All other augmentations are applied to both target and noise and then the
energy levels are measured and the appropriate noise gain calculated to
achieve the desired SNR.

Special values:

 -99    Noise only mixture (no target)
 99     Target only mixture (no noise)
""" + default
    # fmt: on


def doc_random_snrs() -> str:
    default = f"\nDefault value: {get_default_config()['random_snrs']}"
    # fmt: off
    return """
'random_snrs' is a mixture database configuration parameter that specifies a
list of random signal-to-noise ratios. The value(s) must be specified as
random using the syntax: 'rand(<min>, <max>)'.

Random SNRs behave slightly differently from regular or ordered SNRs. As with
ordered SNRs, all other augmentations are applied to both target and noise and
then the energy levels are measured and the appropriate noise gain calculated
to achieve the desired SNR. However, unlike ordered SNRs, the desired SNR is
randomized (per the given rule(s)) for each mixture, i.e., previous random
SNRs are not saved and reused.
""" + default
    # fmt: on


def doc_noise_mix_mode() -> str:
    default = f"\nDefault value: {get_default_config()['noise_mix_mode']}"
    # fmt: off
    return """
'noise_mix_mode' is a mixture database configuration parameter that sets
how to mix noises with targets.

Supported modes:

 exhaustive          Use every noise/augmentation with every target/augmentation.
 non-exhaustive      Cycle through every target/augmentation without necessarily
                     using all noise/augmentation combinations (reduced data set).
 non-combinatorial   Combine a target/augmentation with a single cut of a
                     noise/augmentation non-exhaustively (each target/augmentation
                     does not use each noise/augmentation). Cut has a random start
                     and loops back to the beginning if the end of a
                     noise/augmentation is reached.
""" + default
    # fmt: on


def doc_impulse_responses() -> str:
    default = f"\nDefault value: {get_default_config()['impulse_responses']}"
    # fmt: off
    return """
'impulse_responses' is a mixture database configuration parameter that specifies a
list of impulse response files to use.

See 'augmentations' for details.
""" + default
    # fmt: on


def doc_spectral_masks() -> str:
    default = f"\nDefault value: {get_default_config()['spectral_masks']}"
    # fmt: off
    return """
'spectral_masks' is a mixture database configuration parameter that specifies
a list of spectral mask rules.

All other augmentations are applied including SNR and a mixture is generated
and then the spectral mask rules are applied to the resulting mixture feature.

Rules must specify all the following parameters:

 'f_max_width'      Frequency mask maximum width in bins
 'f_num'            Number of frequency masks to apply (set to 0 to apply none)
 't_max_width'      Time mask maximum width in frames
 't_num'            Number of time masks to apply (set to 0 to apply none)
 't_max_percent'    Upper bound on the width of the time mask in percent
""" + default
    # fmt: on


def doc_config() -> str:
    from ..mixture.constants import VALID_CONFIGS

    text = "\n"
    text += "The SonusAI database is defined using a config.yml file.\n\n"
    text += "See the following for details:\n\n"
    for c in VALID_CONFIGS:
        text += f" {c}\n"
    return text


def doc_asr_configs() -> str:
    from ..utils.asr import get_available_engines

    default = f"\nDefault value: {get_default_config()['asr_configs']}"
    engines = get_available_engines()
    # fmt: off
    text = """
'asr_configs' is a mixture database configuration parameter that sets the list of
ASR engine(s) to use.

Required fields:

 'name'         Unique identifier for the ASR engine.
 'engine'       ASR engine to use. Available engines:
"""
    text += f"                {', '.join(engines)}\n"
    text += """
Optional fields:

 'model'        Some ASR engines allow the specification of a model, but note most are
                very computationally demanding and can overwhelm/hang a local system.
                Available whisper ASR engines:
                    tiny.en, tiny, base.en, base, small.en, small, medium.en, medium, large-v1, large-v2, large
 'device'       Some ASR engines allow the specification of a device, either 'cpu' or 'cuda'.
 'cpu_threads'  Some ASR engines allow the specification of the number of CPU threads to use.
 'compute_type' Some ASR engines allow the specification of a compute type, e.g. 'int8'.
 'beam_size'    Some ASR engines allow the specification of a beam size.
 <other>        Other parameters can be injected into the ASR engine as needed; all
                fields in each config are forwarded to the given engine.

Example:

asr_configs:
  - name: faster_tiny_cuda
    engine: faster_whisper
    model: tiny
    device: cuda
    beam_size: 5
  - name: google
    engine: google

Creates two ASR engines for use named faster_tiny_cuda and google.
"""
    # fmt: on
    return text + default



import logging
from importlib import metadata
from os.path import dirname

from rich.logging import RichHandler

__version__ = metadata.version(__package__)  # pyright: ignore [reportArgumentType]
BASEDIR = dirname(__file__)

commands_doc = """
   audiofe                      Audio front end
   calc_metric_spenh            Run speech enhancement and analysis
   doc                          Documentation
   genft                        Generate feature and truth data
   genmetrics                   Generate mixture metrics data
   genmix                       Generate mixture and truth data
   genmixdb                     Generate a mixture database
   lsdb                         List information about a mixture database
   mkwav                        Make WAV files from a mixture database
   onnx_predict                 Run ONNX predict on a trained model
   summarize_metric_spenh       Summarize speech enhancement and analysis results
   vars                         List custom SonusAI variables
"""

# create logger
logger = logging.getLogger("sonusai")
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(message)s")
formatter_db = logging.Formatter("%(asctime)s %(message)s")
console_handler = RichHandler(show_level=False, show_path=False, show_time=False)
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)

logger_db = logging.getLogger("sonusai_db")
logger_db.setLevel(logging.DEBUG)

# create file handler
def create_file_handler(filename: str) -> None:
    from pathlib import Path

    fh = logging.FileHandler(filename=filename, mode="w")
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    filename_db = Path(filename)
    filename_db = filename_db.parent / (filename_db.stem + "_dbtrace" + filename_db.suffix)
    fh = logging.FileHandler(filename=filename_db, mode="w")
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter_db)
    logger_db.addHandler(fh)


# update console handler
def update_console_handler(verbose: bool) -> None:
    if not verbose:
        logger.removeHandler(console_handler)
        console_handler.setLevel(logging.INFO)
        logger.addHandler(console_handler)


# write initial log message
def initial_log_messages(name: str, subprocess: str | None = None) -> None:
    from datetime import datetime
    from getpass import getuser
    from os import getcwd
    from socket import gethostname
    from sys import argv

    if subprocess is None:
        logger.info(f"SonusAI {__version__}")
    else:
        logger.info(f"SonusAI {subprocess}")
    logger.info(f"{name}")
    logger.info("")
    logger.debug(f"Host:      {gethostname()}")
    logger.debug(f"User:      {getuser()}")
    logger.debug(f"Directory: {getcwd()}")
    logger.debug(f"Date:      {datetime.now()}")
    logger.debug(f"Command:   {' '.join(argv)}")
    logger.debug("")


def commands_list(doc: str = commands_doc) -> list[str]:
    lines = doc.split("\n")
    commands = []
    for line in lines:
        command = line.strip().split(" ").pop(0)
        if command:
            commands.append(command)
    return commands


def exception_handler(e: Exception) -> None:
    import sys

    from rich.console import Console

    logger.error(f"{type(e).__name__}: {e}")
    handlers = [handler for handler in logger.handlers if isinstance(handler, logging.FileHandler)]
    logger.error(f"See {', '.join(handler.baseFilename for handler in handlers)} for details")

    console = Console(color_system=None)
    with console.capture() as capture:
        console.print_exception(show_locals=False)
    logger.debug(capture.get())
    sys.exit(1)



"""sonusai summarize_metric_spenh

usage: summarize_metric_spenh [-hr] [-s SORT] LOC

options:
    -h, --help
    -s SORT, --sort SORT            Sort by SORT column. [default: MIXID]
    -r, --reverse                   Reverse sort order.

Summarize speech enhancement metrics results using data generated by SonusAI calc_metric_spenh.

Inputs:
    LOC         A SonusAI calc_metric_spenh results directory.

"""


def summarize_metric_spenh(location: str, by: str = "MIXID", reverse: bool = False) -> str:
    import glob

    import pandas as pd

    files = sorted(glob.glob(location + "/*_metric_spenh.txt"))
    need_header = True
    header = ["MIXID"]
    data = []
    for file in files:
        with open(file) as f:
            for i, line in enumerate(f):
                if i == 1 and need_header:
                    need_header = False
                    header.extend(line.strip().split())
                elif i == 2:
                    data.append(line.strip().split())
                    break

    df = pd.DataFrame(data, columns=header)  # pyright: ignore [reportArgumentType]
    df[header[0:-2]] = df[header[0:-2]].apply(pd.to_numeric, errors="coerce")
    return df.sort_values(by=by, ascending=not reverse).to_string(index=False)


def main():
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    by = args["--sort"]
    reverse = args["--reverse"]
    location = args["LOC"]

    print(summarize_metric_spenh(location, by, reverse))


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)



"""sonusai calc_metric_spenh

usage: calc_metric_spenh [-hvtpws] [-i MIXID] [-e ASR] [-n NCPU] PLOC TLOC

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to process, can be range like 0:maxmix+1. [default: *]
    -t, --truth-est-mode        Calculate extraction and metrics using truth (instead of prediction).
    -p, --plot                  Enable PDF plots file generation per mixture.
    -w, --wav                   Generate WAV files per mixture.
    -s, --summary               Enable summary files generation.
    -n, --num_process NCPU      Number of parallel processes to use [default: auto]
    -e ASR, --asr-method ASR    ASR method used for WER metrics.  Must exist in the TLOC dataset as pre-calculated
                                metrics using SonusAI genmetrics. Can be either an integer index, i.e 0,1,... or the
                                name of the asr_engine configuration in the dataset.  If an incorrect name is specified,
                                a list of asr_engines of the dataset will be printed.

Calculate speech enhancement metrics of prediction data in PLOC using SonusAI mixture data in TLOC as truth/label
reference. Metric and extraction data files are written into PLOC.

PLOC  directory containing prediction data in .h5 files created from truth/label mixture data in TLOC
TLOC  directory with SonusAI mixture database of truth/label mixture data

For ASR methods, the method must bel2 defined in the TLOC dataset, for example possible fast_whisper available models are:
{tiny.en,tiny,base.en,base,small.en,small,medium.en,medium,large-v1,large-v2,large} and an example configuration looks like:
{'fwhsptiny_cpu': {'engine': 'faster_whisper',
  'model': 'tiny',
  'device': 'cpu',
  'beam_size': 5}}
Note: the ASR config can optionally include the model, device, and other fields the engine supports.
Most ASR are very computationally demanding and can overwhelm/hang a local system.

Outputs the following to PLOC (where id is mixid number 0:num_mixtures):
    <id>_metric_spenh.txt

    If --plot:
        <id>_metric_spenh.pdf

    If --wav:
        <id>_target.wav
        <id>_target_est.wav
        <id>_noise.wav
        <id>_noise_est.wav
        <id>_mixture.wav

        If --truth-est-mode:
            <id>_target_truth_est.wav
            <id>_noise_truth_est.wav

    If --summary:
        metric_spenh_targetf_summary.txt
        metric_spenh_targetf_summary.csv
        metric_spenh_targetf_list.csv
        metric_spenh_targetf_estats_list.csv

        If --truth-est-mode:
            metric_spenh_targetf_truth_list.csv
            metric_spenh_targetf_estats_truth_list.csv

TBD
Metric and extraction data are written into prediction location PLOC as separate files per mixture.

    -d PLOC, --ploc PLOC        Location of SonusAI predict data.

Inputs:

"""

from typing import Any

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from sonusai.datatypes import AudioF
from sonusai.datatypes import AudioT
from sonusai.datatypes import Feature
from sonusai.datatypes import Predict
from sonusai.mixture import MixtureDatabase

DB_99 = np.power(10, 99 / 10)
DB_N99 = np.power(10, -99 / 10)


matplotlib.use("SVG")


def first_key(x: dict) -> str:
    for key in x:
        return key
    raise KeyError("No key found")


def mean_square_error(
    hypothesis: np.ndarray,
    reference: np.ndarray,
    squared: bool = False,
) -> tuple[float, np.ndarray, np.ndarray]:
    """Calculate root-mean-square error or mean square error

    :param hypothesis: [frames, bins]
    :param reference: [frames, bins]
    :param squared: calculate mean square rather than root-mean-square
    :return: mean, mean per bin, mean per frame
    """
    sq_err = np.square(reference - hypothesis)

    # mean over frames for value per bin
    err_b = np.mean(sq_err, axis=0)
    # mean over bins for value per frame
    err_f = np.mean(sq_err, axis=1)
    # mean over all
    err = float(np.mean(sq_err))

    if not squared:
        err_b = np.sqrt(err_b)
        err_f = np.sqrt(err_f)
        err = np.sqrt(err)

    return err, err_b, err_f


def mean_abs_percentage_error(hypothesis: np.ndarray, reference: np.ndarray) -> tuple[float, np.ndarray, np.ndarray]:
    """Calculate mean abs percentage error

    If inputs are complex, calculates average: mape(real)/2 + mape(imag)/2

    :param hypothesis: [frames, bins]
    :param reference: [frames, bins]
    :return: mean, mean per bin, mean per frame
    """
    if not np.iscomplexobj(reference) and not np.iscomplexobj(hypothesis):
        abs_err = 100 * np.abs((reference - hypothesis) / (reference + np.finfo(np.float32).eps))
    else:
        reference_r = np.real(reference)
        reference_i = np.imag(reference)
        hypothesis_r = np.real(hypothesis)
        hypothesis_i = np.imag(hypothesis)
        abs_err_r = 100 * np.abs((reference_r - hypothesis_r) / (reference_r + np.finfo(np.float32).eps))
        abs_err_i = 100 * np.abs((reference_i - hypothesis_i) / (reference_i + np.finfo(np.float32).eps))
        abs_err = (abs_err_r / 2) + (abs_err_i / 2)

    # mean over frames for value per bin
    err_b = np.around(np.mean(abs_err, axis=0), 3)
    # mean over bins for value per frame
    err_f = np.around(np.mean(abs_err, axis=1), 3)
    # mean over all
    err = float(np.around(np.mean(abs_err), 3))

    return err, err_b, err_f


def log_error(reference: np.ndarray, hypothesis: np.ndarray) -> tuple[float, np.ndarray, np.ndarray]:
    """Calculate log error

    :param reference: complex or real [frames, bins]
    :param hypothesis: complex or real [frames, bins]
    :return: mean, mean per bin, mean per frame
    """
    reference_sq = np.real(reference * np.conjugate(reference))
    hypothesis_sq = np.real(hypothesis * np.conjugate(hypothesis))
    log_err = abs(10 * np.log10((reference_sq + np.finfo(np.float32).eps) / (hypothesis_sq + np.finfo(np.float32).eps)))
    # log_err = abs(10 * np.log10(reference_sq / (hypothesis_sq + np.finfo(np.float32).eps) + np.finfo(np.float32).eps))

    # mean over frames for value per bin
    err_b = np.around(np.mean(log_err, axis=0), 3)
    # mean over bins for value per frame
    err_f = np.around(np.mean(log_err, axis=1), 3)
    # mean over all
    err = float(np.around(np.mean(log_err), 3))

    return err, err_b, err_f


def plot_mixpred(
    mixture: AudioT,
    mixture_f: AudioF,
    target: AudioT | None = None,
    feature: Feature | None = None,
    predict: Predict | None = None,
    tp_title: str = "",
) -> tuple[plt.Figure, Any]:  # pyright: ignore [reportPrivateImportUsage]
    from sonusai.constants import SAMPLE_RATE

    num_plots = 2
    if feature is not None:
        num_plots += 1
    if predict is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    p = 0
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[p].plot(x_axis, mixture, label="Mixture", color="mistyrose")
    ax[0].set_ylabel("magnitude", color="tab:blue")
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if target is not None:  # Plot target time-domain waveform on top of mixture
        ax[0].plot(x_axis, target, label="Target", color="tab:blue")
    ax[p].set_title("Waveform")

    # Plot the mixture spectrogram
    p += 1
    ax[p].imshow(np.transpose(mixture_f), aspect="auto", interpolation="nearest", origin="lower")
    ax[p].set_title("Mixture")

    if feature is not None:
        p += 1
        ax[p].imshow(np.transpose(feature), aspect="auto", interpolation="nearest", origin="lower")
        ax[p].set_title("Feature")

    if predict is not None:
        p += 1
        im = ax[p].imshow(np.transpose(predict), aspect="auto", interpolation="nearest", origin="lower")
        ax[p].set_title("Predict " + tp_title)
        plt.colorbar(im, location="bottom")

    return fig, ax


def plot_pdb_predict_truth(
    predict: np.ndarray,
    truth_f: np.ndarray | None = None,
    metric: np.ndarray | None = None,
    tp_title: str = "",
) -> plt.Figure:  # pyright: ignore [reportPrivateImportUsage]
    """Plot predict and optionally truth and a metric in power db, e.g. applies 10*log10(predict)"""
    num_plots = 2
    if truth_f is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    tmp = 10 * np.log10(predict.transpose() + np.finfo(np.float32).eps)
    im = ax[p].imshow(tmp, aspect="auto", interpolation="nearest", origin="lower")
    ax[p].set_title("Predict")
    plt.colorbar(im, location="bottom")

    if truth_f is not None:
        p += 1
        tmp = 10 * np.log10(truth_f.transpose() + np.finfo(np.float32).eps)
        im = ax[p].imshow(tmp, aspect="auto", interpolation="nearest", origin="lower")
        ax[p].set_title("Truth")
        plt.colorbar(im, location="bottom")

    # Plot the predict avg, and optionally truth avg and metric lines
    pred_avg = 10 * np.log10(np.mean(predict, axis=-1) + np.finfo(np.float32).eps)
    p += 1
    x_axis = np.arange(len(pred_avg), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, pred_avg, color="black", linestyle="dashed", label="Predict mean over freq.")
    ax[p].set_ylabel("mean db", color="black")
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_f is not None:
        truth_avg = 10 * np.log10(np.mean(truth_f, axis=-1) + np.finfo(np.float32).eps)
        ax[p].plot(x_axis, truth_avg, color="green", linestyle="dashed", label="Truth mean over freq.")

    if metric is not None:  # instantiate 2nd y-axis that shares the same x-axis
        ax2 = ax[p].twinx()
        color2 = "red"
        ax2.plot(x_axis, metric, color=color2, label="sig distortion (mse db)")
        ax2.set_xlim(x_axis[0], x_axis[-1])
        ax2.set_ylim([0, np.max(metric)])
        ax2.set_ylabel("spectral distortion (mse db)", color=color2)
        ax2.tick_params(axis="y", labelcolor=color2)
        ax[p].set_title("SNR and SNR mse (mean over freq. db)")
    else:
        ax[p].set_title("SNR (mean over freq. db)")
    return fig


def plot_e_predict_truth(
    predict: np.ndarray,
    predict_wav: np.ndarray,
    truth_f: np.ndarray | None = None,
    truth_wav: np.ndarray | None = None,
    metric: np.ndarray | None = None,
    tp_title: str = "",
) -> tuple[plt.Figure, Any]:  # pyright: ignore [reportPrivateImportUsage]
    """Plot predict spectrogram and waveform and optionally truth and a metric)"""
    num_plots = 2
    if truth_f is not None:
        num_plots += 1
    if metric is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    im = ax[p].imshow(predict.transpose(), aspect="auto", interpolation="nearest", origin="lower")
    ax[p].set_title("Predict")
    plt.colorbar(im, location="bottom")

    if truth_f is not None:  # plot truth if provided and use same colormap as predict
        p += 1
        ax[p].imshow(truth_f.transpose(), im.cmap, aspect="auto", interpolation="nearest", origin="lower")
        ax[p].set_title("Truth")

    # Plot predict wav, and optionally truth avg and metric lines
    p += 1
    x_axis = np.arange(len(predict_wav), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, predict_wav, color="black", linestyle="dashed", label="Speech Estimate")
    ax[p].set_ylabel("Amplitude", color="black")
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_wav is not None:
        ntrim = len(truth_wav) - len(predict_wav)
        if ntrim > 0:
            truth_wav = truth_wav[0:-ntrim]
        ax[p].plot(x_axis, truth_wav, color="green", linestyle="dashed", label="True Target")

    # Plot the metric lines
    if metric is not None:
        p += 1
        if metric.ndim > 1:  # if it has multiple dims, plot 1st
            metric1 = metric[:, 0]
        else:
            metric1 = metric  # if single dim, plot it as 1st
        x_axis = np.arange(len(metric1), dtype=np.float32)  # / SAMPLE_RATE
        ax[p].plot(x_axis, metric1, color="red", label="Target LogErr")
        ax[p].set_ylabel("log error db", color="red")
        ax[p].set_xlim(x_axis[0], x_axis[-1])
        ax[p].set_ylim([-0.01, np.max(metric1) + 0.01])
        if metric.ndim > 1 and metric.shape[1] > 1:
            p += 1
            metr2 = metric[:, 1]
            ax = np.append(ax, np.array(ax[p - 1].twinx()))
            color2 = "blue"
            ax[p].plot(x_axis, metr2, color=color2, label="phase dist (deg)")
            # ax2.set_ylim([-180.0, +180.0])
            if np.max(metr2) - np.min(metr2) > 0.1:
                ax[p].set_ylim([np.min(metr2), np.max(metr2)])
            ax[p].set_ylabel("phase dist (deg)", color=color2)
            ax[p].tick_params(axis="y", labelcolor=color2)
            # ax[p].set_title('SNR and SNR mse (mean over freq. db)')

    return fig, ax


def _process_mixture(
    m_id: int,
    truth_location: str,
    predict_location: str,
    predict_wav_mode: bool,
    truth_est_mode: bool,
    enable_plot: bool,
    enable_wav: bool,
    asr_method: str,
    target_f_key: str,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    import pickle
    from os.path import basename
    from os.path import join
    from os.path import splitext

    import h5py
    import pgzip
    from matplotlib.backends.backend_pdf import PdfPages
    from pystoi import stoi

    from sonusai import logger
    from sonusai.metrics import calc_pcm
    from sonusai.metrics import calc_pesq
    from sonusai.metrics import calc_phase_distance
    from sonusai.metrics import calc_speech
    from sonusai.metrics import calc_wer
    from sonusai.metrics import calc_wsdr
    from sonusai.mixture import forward_transform
    from sonusai.mixture import inverse_transform
    from sonusai.mixture import read_audio
    from sonusai.utils import calc_asr
    from sonusai.utils import float_to_int16
    from sonusai.utils import power_compress
    from sonusai.utils import power_uncompress
    from sonusai.utils import reshape_outputs
    from sonusai.utils import stack_complex
    from sonusai.utils import unstack_complex
    from sonusai.utils import write_audio

    mixdb = MixtureDatabase(truth_location)

    # 1)  Read predict data, var predict with shape [BatchSize,Classes] or [batch, timesteps, classes]
    output_name = join(predict_location, mixdb.mixture(m_id).name + ".h5")
    predict = None
    if truth_est_mode:
        # in truth estimation mode we use the truth in place of prediction to see metrics with perfect input
        # don't bother to read prediction, and predict var will get assigned to truth later
        # mark outputs with tru suffix, i.e. 0000_truest_*
        base_name = splitext(output_name)[0] + "_truest"
    else:
        base_name, ext = splitext(output_name)  # base_name used later
        if not predict_wav_mode:
            try:
                with h5py.File(output_name, "r") as f:
                    predict = np.array(f["predict"])
            except Exception as e:
                raise OSError(f"Error reading {output_name}: {e}") from e
            # reshape to always be [frames, classes] where ndim==3 case frames = batch * timesteps
            if predict.ndim > 2:  # TBD generalize to somehow detect if timestep dim exists, some cases > 2 don't have
                # logger.debug(f'Prediction reshape from {predict.shape} to remove timestep dimension.')
                predict, _ = reshape_outputs(predict=predict, truth=None, timesteps=predict.shape[1])
        else:
            base_name, ext = splitext(output_name)
            predict_name = join(base_name + ".wav")
            audio = read_audio(predict_name, use_cache=True)
            predict = forward_transform(audio, mixdb.ft_config)
            if mixdb.feature[0:1] == "h":
                predict = power_compress(predict)
            predict = stack_complex(predict)

    # 2) Collect true target, noise, mixture data, trim to predict size if needed
    tmp = mixdb.mixture_sources(m_id)  # time-dom augmented targets is list of pre-IR and pre-specaugment targets
    target_f = mixdb.mixture_sources_f(m_id, sources=tmp)["primary"]
    target = tmp["primary"]
    mixture = mixdb.mixture_mixture(m_id)  # note: gives full reverberated/distorted target, but no specaugment
    # noise_wo_dist = mixdb.mixture_noise(mixid)            # noise without specaugment and distortion
    # noise_wo_dist_f = mixdb.mixture_noise_f(mixid, noise=noise_wo_dist)
    noise = mixture - target  # has time-domain distortion (ir,etc.) but does not have specaugment
    # noise_f = mixdb.mixture_noise_f(mixid, noise=noise)
    # note: uses pre-IR, pre-specaug audio
    segsnr_f = mixdb.mixture_metrics(m_id, ["ssnr"])["ssnr"]  # Why [0] removed?
    mixture_f = mixdb.mixture_mixture_f(m_id, mixture=mixture)
    noise_f = mixture_f - target_f  # true noise in freq domain includes specaugment and time-domain ir,distortions
    # segsnr_f = mixdb.mixture_segsnr(mixid, target=target, noise=noise)
    segsnr_f[segsnr_f == np.inf] = DB_99
    # segsnr_f should never be -np.inf
    segsnr_f[segsnr_f == -np.inf] = DB_N99
    # need to use inv-tf to match #samples & latency shift properties of predict inv tf
    target_fi = inverse_transform(target_f, mixdb.it_config)
    noise_fi = inverse_transform(noise_f, mixdb.it_config)
    # mixture_fi = mixdb.inverse_transform(mixture_f)

    # gen feature, truth - note feature only used for plots
    # TODO: parse truth_f for different formats
    feature, truth_all = mixdb.mixture_ft(m_id, mixture_f=mixture_f)
    truth_f = truth_all["primary"][target_f_key]
    if truth_f.ndim > 2:  # note this may not be needed anymore as all target_f truth is 3 dims
        if truth_f.shape[1] != 1:
            logger.info("Error: target_f truth has stride > 1, exiting.")
            raise SystemExit(1)
        else:
            truth_f = truth_f[:, 0, :]  # remove stride dimension

    # ignore mixup
    # for truth_setting in mixdb.target_file(mixdb.mixture(mixid).targets[0].file_id).truth_settings:
    #     if truth_setting.function == 'target_mixture_f':
    #         half = truth_f.shape[-1] // 2
    #         # extract target_f only
    #         truth_f = truth_f[..., :half]

    if not truth_est_mode:
        if predict.shape[0] < target_f.shape[0]:  # target_f, truth_f, mixture_f, etc. same size
            trim_f = target_f.shape[0] - predict.shape[0]
            logger.debug(f"Warning: prediction frames less than mixture, trimming {trim_f} frames from all truth.")
            target_f = target_f[0:-trim_f, :]
            target_fi, _ = inverse_transform(target_f, mixdb.it_config)
            trim_t = target.shape[0] - target_fi.shape[0]
            target = target[0:-trim_t]
            noise_f = noise_f[0:-trim_f, :]
            noise = noise[0:-trim_t]
            mixture_f = mixture_f[0:-trim_f, :]
            mixture = mixture[0:-trim_t]
            truth_f = truth_f[0:-trim_f, :]
        elif predict.shape[0] > target_f.shape[0]:
            logger.debug(
                f"Warning: prediction has more frames than true mixture {predict.shape[0]} vs {truth_f.shape[0]}"
            )
            trim_f = predict.shape[0] - target_f.shape[0]
            predict = predict[0:-trim_f, :]
            # raise SonusAIError(
            #     f'Error: prediction has more frames than true mixture {predict.shape[0]} vs {truth_f.shape[0]}')

    # 3) Extraction - format proper complex and wav estimates and truth (unstack, uncompress, inv tf, etc.)
    if truth_est_mode:
        predict = truth_f  # substitute truth for the prediction (for test/debug)
        predict_complex = unstack_complex(predict)  # unstack
        # if feature has compressed mag and truth does not, compress it
        if mixdb.feature[0:1] == "h" and not first_key(mixdb.category_truth_configs("primary")).startswith(
            "targetcmpr"
        ):
            predict_complex = power_compress(predict_complex)  # from uncompressed truth
    else:
        predict_complex = unstack_complex(predict)

    truth_f_complex = unstack_complex(truth_f)
    if mixdb.feature[0:1] == "h":  # 'hn' or 'ha' or 'hd', etc.:  # if feat has compressed mag
        # estimate noise in uncompressed-mag domain
        noise_est_complex = mixture_f - power_uncompress(predict_complex)
        predict_complex = power_uncompress(predict_complex)  # uncompress if truth is compressed
    else:  # cn, c8, ..
        noise_est_complex = mixture_f - predict_complex

    target_est_wav = inverse_transform(predict_complex, mixdb.it_config)
    noise_est_wav = inverse_transform(noise_est_complex, mixdb.it_config)

    # 4) Metrics
    # Target/Speech logerr - PSD estimation accuracy symmetric mean log-spectral distortion
    lerr_tg, lerr_tg_bin, lerr_tg_frame = log_error(reference=truth_f_complex, hypothesis=predict_complex)
    # Noise logerr - PSD estimation accuracy
    lerr_n, lerr_n_bin, lerr_n_frame = log_error(reference=noise_f, hypothesis=noise_est_complex)
    # PCM loss metric
    ytrue_f = np.concatenate((truth_f_complex[:, np.newaxis, :], noise_f[:, np.newaxis, :]), axis=1)
    ypred_f = np.concatenate((predict_complex[:, np.newaxis, :], noise_est_complex[:, np.newaxis, :]), axis=1)
    pcm, pcm_bin, pcm_frame = calc_pcm(hypothesis=ypred_f, reference=ytrue_f, with_log=True)

    # Phase distance
    phd, phd_bin, phd_frame = calc_phase_distance(hypothesis=predict_complex, reference=truth_f_complex)

    # Noise td logerr
    # lerr_nt, lerr_nt_bin, lerr_nt_frame = log_error(noise_fi, noise_truth_est_audio)

    # # SA-SDR (time-domain source-aggregated SDR)
    ytrue = np.concatenate((target_fi[:, np.newaxis], noise_fi[:, np.newaxis]), axis=1)
    ypred = np.concatenate((target_est_wav[:, np.newaxis], noise_est_wav[:, np.newaxis]), axis=1)
    # # note: w/o scale is more pessimistic number
    # sa_sdr, _ = calc_sa_sdr(hypothesis=ypred, reference=ytrue)
    target_stoi = stoi(target_fi, target_est_wav, 16000, extended=False)

    wsdr, wsdr_cc, wsdr_cw = calc_wsdr(hypothesis=ypred, reference=ytrue, with_log=True)
    # logger.debug(f'wsdr weight sum for mixid {mixid} = {np.sum(wsdr_cw)}.')
    # logger.debug(f'wsdr cweights = {wsdr_cw}.')
    # logger.debug(f'wsdr ccoefs for mixid {mixid} = {wsdr_cc}.')

    # Speech intelligibility measure - PESQ
    if int(mixdb.mixture(m_id).noise.snr) > -99:
        # len = target_est_wav.shape[0]
        pesq_speech = calc_pesq(target_est_wav, target_fi)
        csig_tg, cbak_tg, covl_tg = calc_speech(target_est_wav, target_fi, pesq=pesq_speech)
        metrics = mixdb.mixture_metrics(m_id, ["mxpesq", "mxcsig", "mxcbak", "mxcovl"])
        pesq_mx = metrics["mxpesq"][0] if isinstance(metrics["mxpesq"], list) else metrics["mxpesq"]
        csig_mx = metrics["mxcsig"][0] if isinstance(metrics["mxcsig"], list) else metrics["mxcsig"]
        cbak_mx = metrics["mxcbak"][0] if isinstance(metrics["mxcbak"], list) else metrics["mxcbak"]
        covl_mx = metrics["mxcovl"][0] if isinstance(metrics["mxcovl"], list) else metrics["mxcovl"]
        # pesq_speech_tst = calc_pesq(hypothesis=target_est_wav, reference=target)
        # pesq_mixture_tst = calc_pesq(hypothesis=mixture, reference=target)
        # pesq improvement
        pesq_impr = pesq_speech - pesq_mx
        # pesq improvement %
        pesq_impr_pc = pesq_impr / (pesq_mx + np.finfo(np.float32).eps) * 100
    else:
        pesq_speech = 0
        pesq_mx = 0
        pesq_impr_pc = np.float32(0)
        csig_mx = 0
        csig_tg = 0
        cbak_mx = 0
        cbak_tg = 0
        covl_mx = 0
        covl_tg = 0

    # Calc ASR
    asr_tt = None
    asr_mx = None
    asr_tge = None
    # asr_engines = list(mixdb.asr_configs.keys())
    if asr_method is not None and mixdb.mixture(m_id).noise.snr >= -96:  # noise only, ignore/reset target ASR
        asr_mx_name = f"mxasr.{asr_method}"
        wer_mx_name = f"mxwer.{asr_method}"
        asr_tt_name = f"tasr.{asr_method}"
        metrics = mixdb.mixture_metrics(m_id, [asr_mx_name, wer_mx_name, asr_tt_name])
        asr_mx = metrics[asr_mx_name][0] if isinstance(metrics[asr_mx_name], list) else metrics[asr_mx_name]
        wer_mx = metrics[wer_mx_name][0] if isinstance(metrics[wer_mx_name], list) else metrics[wer_mx_name]
        asr_tt = metrics[asr_tt_name][0] if isinstance(metrics[asr_tt_name], list) else metrics[asr_tt_name]

        if asr_tt:
            noiseadd = None  # TBD add as switch, default -30
            if noiseadd is not None:
                ngain = np.power(10, min(float(noiseadd), 0.0) / 20.0)  # limit to gain <1, convert to float
                tgasr_est_wav = target_est_wav + ngain * noise_est_wav  # add back noise at low level
            else:
                tgasr_est_wav = target_est_wav

            # logger.info(f'Calculating prediction ASR for mixid {mixid}')
            asr_cfg = mixdb.asr_configs[asr_method]
            asr_tge = calc_asr(tgasr_est_wav, **asr_cfg).text
            wer_tge = calc_wer(asr_tge, asr_tt).wer * 100  # target estimate WER
            if wer_mx == 0.0:
                if wer_tge == 0.0:
                    wer_pi = 0.0
                else:
                    wer_pi = -999.0  # instead of -Inf
            else:
                wer_pi = 100 * (wer_mx - wer_tge) / wer_mx
        else:
            logger.warning(f"Warning: mixid {m_id} ASR truth is empty, setting to 0% WER")
            wer_mx = float(0)
            wer_tge = float(0)
            wer_pi = float(0)
    else:
        wer_mx = float("nan")
        wer_tge = float("nan")
        wer_pi = float("nan")

    # 5) Save per mixture metric results
    # Single row in table of scalar metrics per mixture
    mtable1_col = [
        "MXSNR",
        "MXPESQ",
        "PESQ",
        "PESQi%",
        "MXWER",
        "WER",
        "WERi%",
        "WSDR",
        "STOI",
        "PCM",
        "SPLERR",
        "NLERR",
        "PD",
        "MXCSIG",
        "CSIG",
        "MXCBAK",
        "CBAK",
        "MXCOVL",
        "COVL",
        "SPFILE",
        "NFILE",
    ]
    ti = mixdb.mixture(m_id).sources["primary"].file_id
    ni = mixdb.mixture(m_id).noise.file_id
    metr1 = [
        mixdb.mixture(m_id).noise.snr,
        pesq_mx,
        pesq_speech,
        pesq_impr_pc,
        wer_mx,
        wer_tge,
        wer_pi,
        wsdr,
        target_stoi,
        pcm,
        lerr_tg,
        lerr_n,
        phd,
        csig_mx,
        csig_tg,
        cbak_mx,
        cbak_tg,
        covl_mx,
        covl_tg,
        basename(mixdb.source_file(ti).name),
        basename(mixdb.source_file(ni).name),
    ]
    mtab1 = pd.DataFrame([metr1], columns=mtable1_col, index=[m_id])

    # Stats of per frame estimation metrics
    metr2 = pd.DataFrame(
        {"SSNR": segsnr_f, "PCM": pcm_frame, "SLERR": lerr_tg_frame, "NLERR": lerr_n_frame, "SPD": phd_frame}
    )
    metr2 = metr2.describe()  # Use pandas stat function
    # Change SSNR stats to dB, except count.  SSNR is index 0, pandas requires using iloc
    # metr2['SSNR'][1:] = metr2['SSNR'][1:].apply(lambda x: 10 * np.log10(x + 1.01e-10))
    metr2.iloc[1:, 0] = metr2["SSNR"][1:].apply(lambda x: 10 * np.log10(x + 1.01e-10))
    # create a single row in multi-column header
    new_labels = pd.MultiIndex.from_product(
        [metr2.columns, ["Avg", "Min", "Med", "Max", "Std"]], names=["Metric", "Stat"]
    )
    dat1row = metr2.loc[["mean", "min", "50%", "max", "std"], :].T.stack().to_numpy().reshape((1, -1))
    mtab2 = pd.DataFrame(dat1row, index=[m_id], columns=new_labels)
    mtab2.insert(0, "MXSNR", mixdb.mixture(m_id).noise.snr, False)  # add MXSNR as the first metric column

    all_metrics_table_1 = mtab1  # return to be collected by process
    all_metrics_table_2 = mtab2  # return to be collected by process

    if asr_method is None:
        metric_name = base_name + "_metric_spenh.txt"
    else:
        metric_name = base_name + "_metric_spenh_" + asr_method + ".txt"

    with open(metric_name, "w") as f:
        print("Speech enhancement metrics:", file=f)
        print(mtab1.round(2).to_string(float_format=lambda x: f"{x:.2f}"), file=f)
        print("", file=f)
        print(f"Extraction statistics over {mixture_f.shape[0]} frames:", file=f)
        print(metr2.round(2).to_string(float_format=lambda x: f"{x:.2f}"), file=f)
        print("", file=f)
        print(f"Target path: {mixdb.source_file(ti).name}", file=f)
        print(f"Noise path: {mixdb.source_file(ni).name}", file=f)
        if asr_method != "none":
            print(f"ASR method: {asr_method}", file=f)
            print(f"ASR truth:  {asr_tt}", file=f)
            print(f"ASR result for mixture:  {asr_mx}", file=f)
            print(f"ASR result for prediction:  {asr_tge}", file=f)

        print(f"Augmentations: {mixdb.mixture(m_id)}", file=f)

    # 7) write wav files
    if enable_wav:
        write_audio(name=base_name + "_mixture.wav", audio=float_to_int16(mixture))
        write_audio(name=base_name + "_target.wav", audio=float_to_int16(target))
        # write_audio(name=base_name + '_target_fi.wav', audio=float_to_int16(target_fi))
        write_audio(name=base_name + "_noise.wav", audio=float_to_int16(noise))
        write_audio(name=base_name + "_target_est.wav", audio=float_to_int16(target_est_wav))
        write_audio(name=base_name + "_noise_est.wav", audio=float_to_int16(noise_est_wav))

        # debug code to test for perfect reconstruction of the extraction method
        # note both 75% olsa-hanns and 50% olsa-hann modes checked to have perfect reconstruction
        # target_r = mixdb.inverse_transform(target_f)
        # noise_r = mixdb.inverse_transform(noise_f)
        # _write_wav(name=base_name + '_target_r.wav', audio=float_to_int16(target_r))
        # _write_wav(name=base_name + '_noise_r.wav', audio=float_to_int16(noise_r)) # chk perfect rec

    # 8) Write out plot file
    if enable_plot:
        plot_name = base_name + "_metric_spenh.pdf"

        # Reshape feature to eliminate overlap redundancy for easier to understand spectrogram view
        # Original size (frames, stride, num_bands), decimates in stride dimension only if step is > 1
        # Reshape to get frames*decimated_stride, num_bands
        step = int(mixdb.feature_samples / mixdb.feature_step_samples)
        if feature.ndim != 3:
            raise OSError("feature does not have 3 dimensions: frames, stride, num_bands")

        # for feature cn*00n**
        feat_sgram = unstack_complex(feature)
        feat_sgram = 20 * np.log10(abs(feat_sgram) + np.finfo(np.float32).eps)
        feat_sgram = feat_sgram[:, -step:, :]  # decimate,  Fx1xB
        feat_sgram = np.reshape(feat_sgram, (feat_sgram.shape[0] * feat_sgram.shape[1], feat_sgram.shape[2]))

        with PdfPages(plot_name) as pdf:
            # page1 we always have a mixture and prediction, target optional if truth provided
            # For speech enhancement, target_f is definitely included:
            predplot = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            tfunc_name = "target_f"
            # if tfunc_name == 'mapped_snr_f':
            #     # leave as unmapped snr
            #     predplot = predict
            #     tfunc_name = mixdb.target_file(1).truth_settings[0].function
            # elif tfunc_name == 'target_f' or 'target_mixture_f':
            #     predplot = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            # else:
            #     # use dB scale
            #     predplot = 10 * np.log10(predict + np.finfo(np.float32).eps)
            #     tfunc_name = tfunc_name + ' (db)'

            mixspec = 20 * np.log10(abs(mixture_f) + np.finfo(np.float32).eps)
            fig, ax = plot_mixpred(
                mixture=mixture,
                mixture_f=mixspec,
                target=target,
                feature=feat_sgram,
                predict=predplot,
                tp_title=tfunc_name,
            )
            pdf.savefig(fig)
            pickle.dump((fig, ax), pgzip.open(base_name + "_metric_spenh_fig1.pkl.gz", "wb"))

            # ----- page 2, plot unmapped predict, opt truth reconstructed and line plots of mean-over-f
            # pdf.savefig(plot_pdb_predtruth(predict=pred_snr_f, tp_title='predict snr_f (db)'))

            # page 3 speech extraction
            tg_spec = 20 * np.log10(abs(target_f) + np.finfo(np.float32).eps)
            tg_est_spec = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            # n_spec = np.reshape(n_spec,(n_spec.shape[0] * n_spec.shape[1], n_spec.shape[2]))
            fig, ax = plot_e_predict_truth(
                predict=tg_est_spec,
                predict_wav=target_est_wav,
                truth_f=tg_spec,
                truth_wav=target_fi,
                metric=np.vstack((lerr_tg_frame, phd_frame)).T,
                tp_title="speech estimate",
            )
            pdf.savefig(fig)
            pickle.dump((fig, ax), pgzip.open(base_name + "_metric_spenh_fig2.pkl.gz", "wb"))

            # page 4 noise extraction
            n_spec = 20 * np.log10(abs(noise_f) + np.finfo(np.float32).eps)
            n_est_spec = 20 * np.log10(abs(noise_est_complex) + np.finfo(np.float32).eps)
            fig, ax = plot_e_predict_truth(
                predict=n_est_spec,
                predict_wav=noise_est_wav,
                truth_f=n_spec,
                truth_wav=noise_fi,
                metric=lerr_n_frame,
                tp_title="noise estimate",
            )
            pdf.savefig(fig)
            pickle.dump((fig, ax), pgzip.open(base_name + "_metric_spenh_fig4.pkl.gz", "wb"))

            # Plot error waveforms
            # tg_err_wav = target_fi - target_est_wav
            # tg_err_spec = 20*np.log10(np.abs(target_f - predict_complex))

        plt.close("all")

    return all_metrics_table_1, all_metrics_table_2


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args["--verbose"]
    mixids = args["--mixid"]
    asr_method = args["--asr-method"]
    truth_est_mode = args["--truth-est-mode"]
    enable_plot = args["--plot"]
    enable_wav = args["--wav"]
    enable_summary = args["--summary"]
    predict_location = args["PLOC"]
    num_proc = args["--num_process"]
    truth_location = args["TLOC"]

    import glob
    from functools import partial
    from os.path import basename
    from os.path import isdir
    from os.path import join

    import psutil

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import par_track
    from sonusai.utils import track

    # Check prediction subdirectory
    if not isdir(predict_location):
        print(f"The specified predict location {predict_location} is not a valid subdirectory path, exiting.")

    # all_predict_files = listdir(predict_location)
    all_predict_files = glob.glob(predict_location + "/*.h5")
    predict_logfile = glob.glob(predict_location + "/*predict.log")
    predict_wav_mode = False
    if len(all_predict_files) <= 0 and not truth_est_mode:
        all_predict_files = glob.glob(predict_location + "/*.wav")  # check for wav files
        if len(all_predict_files) <= 0:
            print(f"Subdirectory {predict_location} has no .h5 or .wav files, exiting.")
        else:
            logger.info(f"Found {len(all_predict_files)} prediction .wav files.")
            predict_wav_mode = True
    else:
        logger.info(f"Found {len(all_predict_files)} prediction .h5 files.")

    if len(predict_logfile) == 0:
        logger.info(f"Warning, predict location {predict_location} has no prediction log files.")
    else:
        logger.info(f"Found predict log {basename(predict_logfile[0])} in predict location.")

    # Setup logging file
    create_file_handler(join(predict_location, "calc_metric_spenh.log"))
    update_console_handler(verbose)
    initial_log_messages("calc_metric_spenh")

    mixdb = MixtureDatabase(truth_location)
    mixids = mixdb.mixids_to_list(mixids)
    logger.info(
        f"Found mixdb of {mixdb.num_mixtures} total mixtures, with {mixdb.num_classes} classes in {truth_location}"
    )
    # speech enhancement metrics and audio truth requires target_f truth type, check it is present
    target_f_key = None
    logger.info(
        f"mixdb has {len(mixdb.category_truth_configs('primary'))} truth types defined for primary, checking that target_f type is present."
    )
    for key in mixdb.category_truth_configs("primary"):
        if mixdb.category_truth_configs("primary")[key] == "target_f":
            target_f_key = key
    if target_f_key is None:
        logger.error("mixdb does not have target_f truth defined, required for speech enhancement metrics, exiting.")
        raise SystemExit(1)

    logger.info(f"Only running specified subset of {len(mixids)} mixtures")

    asr_config_en = None
    fnb = "metric_spenh_"
    if asr_method is not None:
        if asr_method in mixdb.asr_configs:
            logger.info(f"Specified ASR method {asr_method} exists in mixdb.asr_configs, it will be used for ")
            logger.info("prediction ASR and WER, and pre-calculated target and mixture ASR if available.")
            asr_config_en = True
            asr_cfg = mixdb.asr_configs[asr_method]
            fnb = "metric_spenh_" + asr_method + "_"
            logger.info(f"Using ASR cfg: {asr_cfg} ")
            # audio = read_audio(DEFAULT_SPEECH, use_cache=True)
            # logger.info(f'Warming up {asr_method}, note for cloud service this could take up to a few minutes.')
            # asr_chk = calc_asr(audio, **asr_cfg)
            # logger.info(f'Warmup completed, results {asr_chk}')
        else:
            logger.info(
                f"Specified ASR method {asr_method} does not exists in mixdb.asr_configs."
                f"Must choose one of the following (or none):"
            )
            logger.info(f"{', '.join(mixdb.asr_configs)}")
            logger.error("Unrecognized ASR method, exiting.")
            raise SystemExit(1)

    num_cpu = psutil.cpu_count()
    cpu_percent = psutil.cpu_percent(interval=1)
    logger.info(f"#CPUs: {num_cpu}, current CPU utilization: {cpu_percent}%")
    logger.info(f"Memory utilization: {psutil.virtual_memory().percent}%")
    if num_proc == "auto":
        use_cpu = int(num_cpu * (0.9 - cpu_percent / 100))  # default use 80% of available cpus
    elif num_proc == "None":
        use_cpu = None
    else:
        use_cpu = min(max(int(num_proc), 1), num_cpu)

    # Individual mixtures use pandas print, set precision to 2 decimal places
    # pd.set_option('float_format', '{:.2f}'.format)
    logger.info(f"Calculating metrics for {len(mixids)} mixtures using {use_cpu} parallel processes")
    # progress = tqdm(total=len(mixids), desc='calc_metric_spenh', mininterval=1)
    progress = track(total=len(mixids))
    if use_cpu is None:
        no_par = True
        num_cpus = None
    else:
        no_par = False
        num_cpus = use_cpu

    all_metrics_tables = par_track(
        partial(
            _process_mixture,
            truth_location=truth_location,
            predict_location=predict_location,
            predict_wav_mode=predict_wav_mode,
            truth_est_mode=truth_est_mode,
            enable_plot=enable_plot,
            enable_wav=enable_wav,
            asr_method=asr_method,
            target_f_key=target_f_key,
        ),
        mixids,
        progress=progress,
        num_cpus=num_cpus,
        no_par=no_par,
    )
    progress.close()

    all_metrics_table_1 = pd.concat([item[0] for item in all_metrics_tables])
    all_metrics_table_2 = pd.concat([item[1] for item in all_metrics_tables])

    if not enable_summary:
        return

    # 9) Done with mixtures, write out summary metrics
    # Calculate SNR summary avg of each non-random snr
    all_mtab1_sorted = all_metrics_table_1.sort_values(by=["MXSNR", "SPFILE"])
    all_mtab2_sorted = all_metrics_table_2.sort_values(by=["MXSNR"])
    mtab_snr_summary = None
    mtab_snr_summary_em = None
    for snri in range(0, len(mixdb.snrs)):
        tmp = all_mtab1_sorted.query("MXSNR==" + str(mixdb.snrs[snri])).mean(numeric_only=True).to_frame().T
        # avoid nan when subset of mixids specified
        if ~np.isnan(tmp.iloc[0].to_numpy()[0]).any():
            mtab_snr_summary = pd.concat([mtab_snr_summary, tmp])

        tmp = all_mtab2_sorted[all_mtab2_sorted["MXSNR"] == mixdb.snrs[snri]].mean(numeric_only=True).to_frame().T
        # avoid nan when subset of mixids specified (mxsnr will be nan if no data):
        if ~np.isnan(tmp.iloc[0].to_numpy()[0]).any():
            mtab_snr_summary_em = pd.concat([mtab_snr_summary_em, tmp])

    mtab_snr_summary = mtab_snr_summary.sort_values(by=["MXSNR"], ascending=False)
    # Correct percentages in snr summary table
    mtab_snr_summary["PESQi%"] = (
        100 * (mtab_snr_summary["PESQ"] - mtab_snr_summary["MXPESQ"]) / np.maximum(mtab_snr_summary["MXPESQ"], 0.01)
    )
    for i in range(len(mtab_snr_summary)):
        if mtab_snr_summary["MXWER"].iloc[i] == 0.0:
            if mtab_snr_summary["WER"].iloc[i] == 0.0:
                mtab_snr_summary.iloc[i, 6] = 0.0  # mtab_snr_summary['WERi%'].iloc[i] = 0.0
            else:
                mtab_snr_summary.iloc[i, 6] = -999.0  # mtab_snr_summary['WERi%'].iloc[i] = -999.0
        else:
            if ~np.isnan(mtab_snr_summary["WER"].iloc[i]) and ~np.isnan(mtab_snr_summary["MXWER"].iloc[i]):
                # update WERi% in 6th col
                mtab_snr_summary.iloc[i, 6] = (
                    100
                    * (mtab_snr_summary["MXWER"].iloc[i] - mtab_snr_summary["WER"].iloc[i])
                    / mtab_snr_summary["MXWER"].iloc[i]
                )

    # Calculate avg metrics over all mixtures except -99
    all_mtab1_sorted_nom99 = all_mtab1_sorted[all_mtab1_sorted.MXSNR != -99]
    all_nom99_mean = all_mtab1_sorted_nom99.mean(numeric_only=True)

    # correct the percentage averages with a direct calculation (PESQ% and WER%):
    # ser.iloc[pos]
    all_nom99_mean["PESQi%"] = (
        100 * (all_nom99_mean["PESQ"] - all_nom99_mean["MXPESQ"]) / np.maximum(all_nom99_mean["MXPESQ"], 0.01)
    )  # pesq%
    # all_nom99_mean[3] = 100 * (all_nom99_mean[2] - all_nom99_mean[1]) / np.maximum(all_nom99_mean[1], 0.01)  # pesq%
    if all_nom99_mean["MXWER"] == 0.0:
        if all_nom99_mean["WER"] == 0.0:
            all_nom99_mean["WERi%"] = 0.0
        else:
            all_nom99_mean["WERi%"] = -999.0
    else:  # WER%
        all_nom99_mean["WERi%"] = 100 * (all_nom99_mean["MXWER"] - all_nom99_mean["WER"]) / all_nom99_mean["MXWER"]

    num_mix = len(mixids)
    if num_mix > 1:
        # Print pandas data to files using precision to 2 decimals
        # pd.set_option('float_format', '{:.2f}'.format)

        if not truth_est_mode:
            ofname = join(predict_location, fnb + "summary.txt")
        else:
            ofname = join(predict_location, fnb + "summary_truest.txt")

        with open(ofname, "w") as f:
            print(f"ASR enabled with method {asr_method}", file=f)
            print(
                f"Speech enhancement metrics avg over all {len(all_mtab1_sorted_nom99)} non -99 SNR mixtures:", file=f
            )
            print(
                all_nom99_mean.to_frame().T.round(2).to_string(float_format=lambda x: f"{x:.2f}", index=False), file=f
            )
            print("\nSpeech enhancement metrics avg over each SNR:", file=f)
            print(mtab_snr_summary.round(2).to_string(float_format=lambda x: f"{x:.2f}", index=False), file=f)
            print("", file=f)
            print("Extraction statistics stats avg over each SNR:", file=f)
            # with pd.option_context('display.max_colwidth', 9):
            # with pd.set_option('float_format', '{:.1f}'.format):
            print(mtab_snr_summary_em.round(1).to_string(float_format=lambda x: f"{x:.1f}", index=False), file=f)
            print("", file=f)
            # pd.set_option('float_format', '{:.2f}'.format)

            print(f"Speech enhancement metrics stats over all {num_mix} mixtures:", file=f)
            print(all_metrics_table_1.describe().round(2).to_string(float_format=lambda x: f"{x:.2f}"), file=f)
            print("", file=f)
            print(f"Extraction statistics stats over all {num_mix} mixtures:", file=f)
            print(all_metrics_table_2.describe().round(2).to_string(float_format=lambda x: f"{x:.1f}"), file=f)
            print("", file=f)

            print("Speech enhancement metrics all-mixtures list:", file=f)
            # print(all_metrics_table_1.head().style.format(precision=2), file=f)
            print(all_metrics_table_1.round(2).to_string(float_format=lambda x: f"{x:.2f}"), file=f)
            print("", file=f)
            print("Extraction statistics all-mixtures list:", file=f)
            print(all_metrics_table_2.round(2).to_string(float_format=lambda x: f"{x:.1f}"), file=f)

        # Write summary to .csv file
        if not truth_est_mode:
            csv_name = str(join(predict_location, fnb + "summary.csv"))
        else:
            csv_name = str(join(predict_location, fnb + "truest_summary.csv"))
        header_args = {
            "mode": "a",
            "encoding": "utf-8",
            "index": False,
            "header": False,
        }
        table_args = {
            "mode": "a",
            "encoding": "utf-8",
        }
        label = f"Speech enhancement metrics avg over all {len(all_mtab1_sorted_nom99)} non -99 SNR mixtures:"
        pd.DataFrame([label]).to_csv(csv_name, header=False, index=False)  # open as write
        all_nom99_mean.to_frame().T.round(2).to_csv(csv_name, index=False, **table_args)
        pd.DataFrame([""]).to_csv(csv_name, **header_args)
        pd.DataFrame(["Speech enhancement metrics avg over each SNR:"]).to_csv(csv_name, **header_args)
        mtab_snr_summary.round(2).to_csv(csv_name, index=False, **table_args)
        pd.DataFrame([""]).to_csv(csv_name, **header_args)
        pd.DataFrame(["Extraction statistics stats avg over each SNR:"]).to_csv(csv_name, **header_args)
        mtab_snr_summary_em.round(2).to_csv(csv_name, index=False, **table_args)
        pd.DataFrame([""]).to_csv(csv_name, **header_args)
        pd.DataFrame([""]).to_csv(csv_name, **header_args)
        label = f"Speech enhancement metrics stats over {num_mix} mixtures:"
        pd.DataFrame([label]).to_csv(csv_name, **header_args)
        all_metrics_table_1.describe().round(2).to_csv(csv_name, **table_args)
        pd.DataFrame([""]).to_csv(csv_name, **header_args)
        label = f"Extraction statistics stats over {num_mix} mixtures:"
        pd.DataFrame([label]).to_csv(csv_name, **header_args)
        all_metrics_table_2.describe().round(2).to_csv(csv_name, **table_args)
        label = f"ASR enabled with method {asr_method}"
        pd.DataFrame([label]).to_csv(csv_name, **header_args)

        if not truth_est_mode:
            csv_name = str(join(predict_location, fnb + "list.csv"))
        else:
            csv_name = str(join(predict_location, fnb + "truest_list.csv"))
        pd.DataFrame(["Speech enhancement metrics list:"]).to_csv(csv_name, header=False, index=False)  # open as write
        all_metrics_table_1.round(2).to_csv(csv_name, **table_args)

        if not truth_est_mode:
            csv_name = str(join(predict_location, fnb + "estats_list.csv"))
        else:
            csv_name = str(join(predict_location, fnb + "truest_estats_list.csv"))
        pd.DataFrame(["Extraction statistics list:"]).to_csv(csv_name, header=False, index=False)  # open as write
        all_metrics_table_2.round(2).to_csv(csv_name, **table_args)


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)

# if asr_method == 'none':
#     fnb = 'metric_spenh_'
# elif asr_method == 'google':
#     fnb = 'metric_spenh_ggl_'
#     logger.info(f'ASR enabled with method {asr_method}')
#     enable_asr_warmup = True
# elif asr_method == 'deepgram':
#     fnb = 'metric_spenh_dgram_'
#     logger.info(f'ASR enabled with method {asr_method}')
#     enable_asr_warmup = True
# elif asr_method == 'aixplain_whisper':
#     fnb = 'metric_spenh_whspx_' + mixdb.asr_configs[asr_method]['model'] + '_'
#     asr_model_name = mixdb.asr_configs[asr_method]['model']
#     enable_asr_warmup = True
# elif asr_method == 'whisper':
#     fnb = 'metric_spenh_whspl_' + mixdb.asr_configs[asr_method]['model'] + '_'
#     asr_model_name = mixdb.asr_configs[asr_method]['model']
#     enable_asr_warmup = True
# elif asr_method == 'aaware_whisper':
#     fnb = 'metric_spenh_whspaaw_' + mixdb.asr_configs[asr_method]['model'] + '_'
#     asr_model_name = mixdb.asr_configs[asr_method]['model']
#     enable_asr_warmup = True
# elif asr_method == 'faster_whisper':
#     fnb = 'metric_spenh_fwhsp_' + mixdb.asr_configs[asr_method]['model'] + '_'
#     asr_model_name = mixdb.asr_configs[asr_method]['model']
#     enable_asr_warmup = True
# elif asr_method == 'sensory':
#     fnb = 'metric_spenh_snsr_' + mixdb.asr_configs[asr_method]['model'] + '_'
#     asr_model_name = mixdb.asr_configs[asr_method]['model']
#     enable_asr_warmup = True
# else:
#     logger.error(f'Unrecognized ASR method: {asr_method}')
#     return



from typing import Any


def import_module(name: str) -> Any:
    """Import a Python module adding the module file's directory to the Python system path so that relative package
    imports are found correctly.
    """
    import os
    import sys
    from importlib import import_module

    try:
        path = os.path.dirname(name)
        if len(path) < 1:
            path = "./"

        # Add model file location to system path
        sys.path.append(os.path.abspath(path))

        try:
            root = os.path.splitext(os.path.basename(name))[0]
            model = import_module(root)
        except Exception as e:
            raise OSError(f"Error: could not import model from {name}: {e}.") from e
    except Exception as e:
        raise OSError(f"Error: could not find {name}: {e}.") from e

    return model



def log_duration_and_sizes(
    total_duration: float,
    feature_step_samples: int,
    feature_parameters: int,
    stride: int,
    desc: str,
) -> None:
    from .. import logger
    from ..constants import FLOAT_BYTES
    from ..constants import SAMPLE_BYTES
    from ..constants import SAMPLE_RATE
    from ..utils.human_readable_size import human_readable_size
    from ..utils.seconds_to_hms import seconds_to_hms

    total_samples = int(total_duration * SAMPLE_RATE)
    mixture_bytes = total_samples * SAMPLE_BYTES
    feature_bytes = total_samples / feature_step_samples * stride * feature_parameters * FLOAT_BYTES

    logger.info("")
    logger.info(f"{desc} duration:   {seconds_to_hms(seconds=total_duration)}")
    logger.info(f"{desc} sizes:")
    logger.info(f" mixture:             {human_readable_size(mixture_bytes, 1)}")
    logger.info(f" feature:             {human_readable_size(feature_bytes, 1)}")



"""sonusai mkwav

usage: mkwav [-hvtsn] [-i MIXID] LOC

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -t, --source                    Write source file.
    -s, --sources                   Write sources files.
    -n, --noise                     Write noise file.

The mkwav command creates WAV files from a SonusAI database.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.

Outputs the following to the mixture database directory:
    <id>
        mixture.wav:        mixture
        source.wav:         source (optional)
        source_<c>.wav:     source <category> (optional)
        noise.wav:          noise (optional)
        metadata.txt
    mkwav.log

"""


def _process_mixture(m_id: int, location: str, write_target: bool, write_targets: bool, write_noise: bool) -> None:
    from os import makedirs
    from os.path import join

    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import write_mixture_metadata
    from sonusai.utils import float_to_int16
    from sonusai.utils import write_audio

    mixdb = MixtureDatabase(location)

    index = mixdb.mixture(m_id).name
    location = join(mixdb.location, "mixture", index)
    makedirs(location, exist_ok=True)

    write_audio(name=join(location, "mixture.wav"), audio=float_to_int16(mixdb.mixture_mixture(m_id)))
    if write_target:
        write_audio(name=join(location, "source.wav"), audio=float_to_int16(mixdb.mixture_source(m_id)))
    if write_targets:
        for category, source in mixdb.mixture_sources(m_id).items():
            write_audio(name=join(location, f"sources_{category}.wav"), audio=float_to_int16(source))
    if write_noise:
        write_audio(name=join(location, "noise.wav"), audio=float_to_int16(mixdb.mixture_noise(m_id)))

    write_mixture_metadata(mixdb, m_id=m_id)


def main() -> None:
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    verbose = args["--verbose"]
    mixid = args["--mixid"]
    write_source = args["--source"]
    write_sources = args["--sources"]
    write_noise = args["--noise"]
    location = args["LOC"]

    import time
    from functools import partial
    from os.path import join

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import check_audio_files_exist
    from sonusai.utils import par_track
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import track

    start_time = time.monotonic()

    create_file_handler(join(location, "mkwav.log"))
    update_console_handler(verbose)
    initial_log_messages("mkwav")

    logger.info(f"Load mixture database from {location}")
    mixdb = MixtureDatabase(location)
    mixid = mixdb.mixids_to_list(mixid)

    total_samples = mixdb.total_samples(mixid)

    logger.info("")
    logger.info(f"Found {len(mixid):,} mixtures to process")
    logger.info(f"{total_samples:,} samples")

    check_audio_files_exist(mixdb)

    progress = track(total=len(mixid))
    par_track(
        partial(
            _process_mixture,
            location=location,
            write_target=write_source,
            write_targets=write_sources,
            write_noise=write_noise,
        ),
        mixid,
        progress=progress,
        # no_par=True,
    )
    progress.close()

    logger.info(f"Wrote {len(mixid)} mixtures to {location}")
    logger.info("")
    end_time = time.monotonic()
    logger.info(f"Completed in {seconds_to_hms(seconds=end_time - start_time)}")
    logger.info("")


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)



import os
from pathlib import Path

from .types import TimeAlignedType


def load_text(audio: str | os.PathLike[str]) -> TimeAlignedType | None:
    """Load time-aligned text data given a TIMIT audio file.

    :param audio: Path to the TIMIT audio file.
    :return: A TimeAlignedType object.
    """
    import string

    from ..mixture.audio import get_sample_rate

    file = Path(audio).with_suffix(".TXT")
    if not os.path.exists(file):
        return None

    sample_rate = get_sample_rate(str(audio))

    with open(file, encoding="utf-8") as f:
        line = f.read()

    fields = line.strip().split()
    start = int(fields[0]) / sample_rate
    end = int(fields[1]) / sample_rate
    text = " ".join(fields[2:]).lower().translate(str.maketrans("", "", string.punctuation))

    return TimeAlignedType(start, end, text)


def load_words(audio: str | os.PathLike[str]) -> list[TimeAlignedType] | None:
    """Load time-aligned word data given a TIMIT audio file.

    :param audio: Path to the TIMIT audio file.
    :return: A list of TimeAlignedType objects.
    """

    return _load_ta(audio, "words")


def load_phonemes(audio: str | os.PathLike[str]) -> list[TimeAlignedType] | None:
    """Load time-aligned phonemes data given a TIMIT audio file.

    :param audio: Path to the TIMIT audio file.
    :return: A list of TimeAlignedType objects.
    """

    return _load_ta(audio, "phonemes")


def _load_ta(audio: str | os.PathLike[str], tier: str) -> list[TimeAlignedType] | None:
    from ..mixture.audio import get_sample_rate

    if tier == "words":
        file = Path(audio).with_suffix(".WRD")
    elif tier == "phonemes":
        file = Path(audio).with_suffix(".PHN")
    else:
        raise ValueError(f"Unknown tier: {tier}")

    if not os.path.exists(file):
        return None

    sample_rate = get_sample_rate(str(audio))

    entries: list[TimeAlignedType] = []
    first = True
    with open(file, encoding="utf-8") as f:
        for line in f.readlines():
            fields = line.strip().split()
            start = int(fields[0]) / sample_rate
            end = int(fields[1]) / sample_rate
            text = " ".join(fields[2:])

            if first:
                first = False
            else:
                if start < entries[-1].end:
                    start = entries[-1].end - (entries[-1].end - start) // 2
                    entries[-1] = TimeAlignedType(text=entries[-1].text, start=entries[-1].start, end=start)

                if end <= start:
                    end = start + 1 / sample_rate

            entries.append(TimeAlignedType(text=text, start=start, end=end))

    return entries


def _years_between(record, born):
    try:
        rec_fields = [int(x) for x in record.split("/")]
        brn_fields = [int(x) for x in born.split("/")]
        return rec_fields[2] - brn_fields[2] - ((rec_fields[1], rec_fields[0]) < (brn_fields[1], brn_fields[0]))
    except ValueError:
        return "??"


def _decode_dialect(d: str) -> str:
    if d in ["DR1", "1"]:
        return "New England"
    if d in ["DR2", "2"]:
        return "Northern"
    if d in ["DR3", "3"]:
        return "North Midland"
    if d in ["DR4", "4"]:
        return "South Midland"
    if d in ["DR5", "5"]:
        return "Southern"
    if d in ["DR6", "6"]:
        return "New York City"
    if d in ["DR7", "7"]:
        return "Western"
    if d in ["DR8", "8"]:
        return "Army Brat"

    raise ValueError(f"Unrecognized dialect: {d}")


def load_speakers(input_dir: Path) -> dict:
    speakers = {}
    with open(input_dir / "SPKRINFO.TXT") as file:
        for line in file:
            if not line.startswith(";"):
                fields = line.strip().split()
                speaker_id = fields[0]
                gender = fields[1]
                dialect = _decode_dialect(fields[2])
                age = _years_between(fields[4], fields[5])
                speakers[speaker_id] = {
                    "gender": gender,
                    "dialect": dialect,
                    "age": age,
                }
    return speakers



# SonusAI general utilities
# ruff: noqa: F401

from .asl_p56 import asl_p56
from .asr import ASRResult
from .asr import calc_asr
from .asr import get_available_engines
from .asr import validate_asr
from .audio_devices import get_default_input_device
from .audio_devices import get_input_device_index_by_name
from .audio_devices import get_input_devices
from .braced_glob import braced_glob
from .braced_glob import braced_iglob
from .calculate_input_shape import calculate_input_shape
from .choice import RandomChoice
from .choice import SequentialChoice
from .compress import power_compress
from .compress import power_uncompress
from .convert_string_to_number import convert_string_to_number
from .create_timestamp import create_timestamp
from .create_ts_name import create_ts_name
from .dataclass_from_dict import dataclass_from_dict
from .db import db_to_linear
from .db import linear_to_db
from .docstring import add_commands_to_docstring
from .docstring import trim_docstring
from .energy_f import compute_energy_f
from .engineering_number import EngineeringNumber
from .evaluate_random_rule import evaluate_random_rule
from .get_frames_per_batch import get_frames_per_batch
from .get_label_names import get_label_names
from .grouper import grouper
from .human_readable_size import human_readable_size
from .keyboard_interrupt import register_keyboard_interrupt
from .load_object import load_object
from .max_text_width import max_text_width
from .model_utils import import_module
from .numeric_conversion import float_to_int16
from .numeric_conversion import int16_to_float
from .onnx_utils import add_sonusai_metadata
from .onnx_utils import get_sonusai_metadata
from .onnx_utils import load_ort_session
from .parallel import par_track
from .parallel import track
from .path_info import PathInfo
from .print_mixture_details import print_class_count
from .print_mixture_details import print_mixture_details
from .rand import seed_context
from .ranges import consolidate_range
from .ranges import expand_range
from .read_predict_data import read_predict_data
from .reshape import get_input_shape
from .reshape import get_num_classes_from_predict
from .reshape import reshape_inputs
from .reshape import reshape_outputs
from .seconds_to_hms import seconds_to_hms
from .stacked_complex import stack_complex
from .stacked_complex import stacked_complex_imag
from .stacked_complex import stacked_complex_real
from .stacked_complex import unstack_complex
from .stratified_shuffle_split import stratified_shuffle_split_mixid
from .tokenized_shell_vars import tokenized_expand
from .tokenized_shell_vars import tokenized_replace
from .write_audio import write_audio
from .yes_or_no import yes_or_no
